"""
==================================
Fiber to bundle coherence measures
==================================

This demo presents the fiber to bundle coherence (FBC) quantitative
measure of the alignment of each fiber with the surrounding fiber bundles
:footcite:p:`Meesters2016b`. These measures are useful in 'cleaning' the results
of tractography algorithms, since low FBCs indicate which fibers are isolated
and poorly aligned with their neighbors, as shown in the figure below.

.. _fiber_to_bundle_coherence:

.. figure:: /_static/images/examples/fbc_illustration.png
   :scale: 60 %
   :align: center

   On the left this figure illustrates (in 2D) the contribution of two fiber
   points to the kernel density estimator. The kernel density estimator is the
   sum over all such locally aligned kernels. The local fiber to bundle
   coherence, shown on the right, color-coded for each fiber, is obtained by
   evaluating the kernel density estimator along the fibers. One spurious
   fiber is present which is isolated and badly aligned with the other fibers,
   and can be identified by a low LFBC value in the region where it deviates
   from the bundle. Figure adapted from :footcite:p:`Portegies2015b`.

Here we implement FBC measures based on kernel density estimation in the
non-flat 5D position-orientation domain. First we compute the kernel density
estimator induced by the full lifted output (defined in the space of positions
and orientations) of the tractography. Then, the Local FBC (LFBC) is the
result of evaluating the estimator along each element of the lifted fiber.
A whole fiber measure, the relative FBC (RFBC), is calculated
by the minimum of the moving average LFBC along the fiber.
Details of the computation of FBC can be found in :footcite:p:`Portegies2015b`.



The FBC measures are evaluated on the Stanford HARDI dataset
(150 orientations, b=2000 $s/mm^2$) which is one of the standard example
datasets in DIPY_.
"""

import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.denoise.enhancement_kernel import EnhancementKernel
from dipy.direction import ProbabilisticDirectionGetter, peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.reconst.csdeconv import ConstrainedSphericalDeconvModel, auto_response_ssst
from dipy.reconst.shm import CsaOdfModel
from dipy.tracking import utils
from dipy.tracking.fbcmeasures import FBCMeasures
from dipy.tracking.local_tracking import LocalTracking
from dipy.tracking.stopping_criterion import ThresholdStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.viz import actor, window

# Enables/disables interactive visualization
interactive = False
# Fix seed
rng = np.random.default_rng(1)

# Read data
hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")
t1_fname = get_fnames(name="stanford_t1")

data, affine = load_nifti(hardi_fname)
labels = load_nifti_data(label_fname)
t1_data = load_nifti_data(t1_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

# Select a relevant part of the data (left hemisphere)
# Coordinates given in x bounds, y bounds, z bounds
dshape = data.shape[:-1]
xa, xb, ya, yb, za, zb = [15, 42, 10, 65, 18, 65]
data_small = data[xa:xb, ya:yb, za:zb]
selectionmask = np.zeros(dshape, "bool")
selectionmask[xa:xb, ya:yb, za:zb] = True

###############################################################################
# The data is first fitted to the Constant Solid Angle (CDA) ODF Model. CSA is
# a good choice to estimate general fractional anisotropy (GFA), which the
# stopping criterion can use to restrict fiber tracking to those areas where
# the ODF shows significant restricted diffusion, thus creating a
# region-of-interest in which the computations are done.

# Perform CSA
csa_model = CsaOdfModel(gtab, sh_order_max=6)
csa_peaks = peaks_from_model(
    csa_model,
    data,
    default_sphere,
    relative_peak_threshold=0.6,
    min_separation_angle=45,
    mask=selectionmask,
)

# Stopping Criterion
stopping_criterion = ThresholdStoppingCriterion(csa_peaks.gfa, 0.25)

###############################################################################
# In order to perform probabilistic fiber tracking we first fit the data to the
# Constrained Spherical Deconvolution (CSD) model in DIPY. This model
# represents each voxel in the data set as a collection of small white matter
# fibers with different orientations. The density of fibers along each
# orientation is known as the Fiber Orientation Distribution (FOD), used in
# the fiber tracking.

# Perform CSD on the original data
response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)
csd_model = ConstrainedSphericalDeconvModel(gtab, response)
csd_fit = csd_model.fit(data_small)
csd_fit_shm = np.pad(
    csd_fit.shm_coeff,
    ((xa, dshape[0] - xb), (ya, dshape[1] - yb), (za, dshape[2] - zb), (0, 0)),
    "constant",
)

# Probabilistic direction getting for fiber tracking
prob_dg = ProbabilisticDirectionGetter.from_shcoeff(
    csd_fit_shm, max_angle=30.0, sphere=default_sphere
)

###############################################################################
# The optic radiation is reconstructed by tracking fibers from the calcarine
# sulcus (visual cortex V1) to the lateral geniculate nucleus (LGN). We seed
# from the calcarine sulcus by selecting a region-of-interest (ROI) cube of
# dimensions 3x3x3 voxels.

# Set a seed region region for tractography.
mask = np.zeros(data.shape[:-1], "bool")
rad = 3
mask[26 - rad : 26 + rad, 29 - rad : 29 + rad, 31 - rad : 31 + rad] = True
seeds = utils.seeds_from_mask(mask, affine, density=[4, 4, 4])

###############################################################################
# Local Tracking is used for probabilistic tractography which takes the
# direction getter along with the stopping criterion and seeds as input.

# Perform tracking using Local Tracking
streamlines_generator = LocalTracking(
    prob_dg, stopping_criterion, seeds, affine, step_size=0.5
)

# Compute streamlines.
streamlines = Streamlines(streamlines_generator)

###############################################################################
# In order to select only the fibers that enter into the LGN, another ROI is
# created from a cube of size 5x5x5 voxels. The near_roi command is used to
# find the fibers that traverse through this ROI.

# Set a mask for the lateral geniculate nucleus (LGN)
mask_lgn = np.zeros(data.shape[:-1], "bool")
rad = 5
mask_lgn[35 - rad : 35 + rad, 42 - rad : 42 + rad, 28 - rad : 28 + rad] = True

# Select all the fibers that enter the LGN and discard all others
filtered_fibers2 = utils.near_roi(streamlines, affine, mask_lgn, tol=1.8)

sfil = []
for i in range(len(streamlines)):
    if filtered_fibers2[i]:
        sfil.append(streamlines[i])
streamlines = Streamlines(sfil)

###############################################################################
# Inspired by :footcite:p:`Rodrigues2010`, a lookup-table is created, containing
# rotated versions of the fiber propagation kernel :math:`P_t`
# :footcite:p:`Duits2011` rotated over a discrete set of orientations. See the
# :ref:`sphx_glr_examples_built_contextual_enhancement_contextual_enhancement.py`
# example for more details regarding the kernel. In order to ensure
# rotationally invariant processing, the discrete orientations are required
# to be equally distributed over a sphere. By default, a sphere with 100
# directions is obtained from electrostatic repulsion in DIPY.

# Compute lookup table
D33 = 1.0
D44 = 0.02
t = 1
k = EnhancementKernel(D33, D44, t)

###############################################################################
# The FBC measures are now computed, taking the tractography results and the
# lookup tables as input.

# Apply FBC measures
fbc = FBCMeasures(streamlines, k)

###############################################################################
# After calculating the FBC measures, a threshold can be chosen on the relative
# FBC (RFBC) in order to remove spurious fibers. Recall that the relative FBC
# (RFBC) is calculated by the minimum of the moving average LFBC along the
# fiber. In this example we show the results for threshold 0 (i.e. all fibers
# are included) and 0.2 (removing the 20 percent most spurious fibers).

# Calculate LFBC for original fibers
fbc_sl_orig, clrs_orig, rfbc_orig = fbc.get_points_rfbc_thresholded(0, emphasis=0.01)

# Apply a threshold on the RFBC to remove spurious fibers
fbc_sl_thres, clrs_thres, rfbc_thres = fbc.get_points_rfbc_thresholded(
    0.125, emphasis=0.01
)

###############################################################################
# The results of FBC measures are visualized, showing the original fibers
# colored by LFBC (see :ref:`this figure <optic_radiation_before_cleaning>`),
# and the fibers after the cleaning procedure via RFBC thresholding (see
# :ref:`this other figure <optic_radiation_after_cleaning>`).

# Create scene
scene = window.Scene()

# Original lines colored by LFBC
lineactor = actor.line(fbc_sl_orig, colors=np.vstack(clrs_orig), linewidth=0.2)
scene.add(lineactor)

# Horizontal (axial) slice of T1 data
vol_actor1 = actor.slicer(t1_data, affine=affine)
vol_actor1.display(z=20)
scene.add(vol_actor1)

# Vertical (sagittal) slice of T1 data
vol_actor2 = actor.slicer(t1_data, affine=affine)
vol_actor2.display(x=35)
scene.add(vol_actor2)

# Show original fibers
scene.set_camera(position=(-264, 285, 155), focal_point=(0, -14, 9), view_up=(0, 0, 1))
window.record(scene=scene, n_frames=1, out_path="OR_before.png", size=(900, 900))
if interactive:
    window.show(scene)

# Show thresholded fibers
scene.rm(lineactor)
scene.add(actor.line(fbc_sl_thres, colors=np.vstack(clrs_thres), linewidth=0.2))
window.record(scene=scene, n_frames=1, out_path="OR_after.png", size=(900, 900))
if interactive:
    window.show(scene)

###############################################################################
# .. _optic_radiation_before_cleaning:
#
# .. rst-class:: centered small fst-italic fw-semibold
#
# The optic radiation obtained through probabilistic tractography colored by
# local fiber to bundle coherence.
#
#
# .. _optic_radiation_after_cleaning:
#
# .. rst-class:: centered small fst-italic fw-semibold
#
# The tractography result is cleaned (shown in bottom) by removing fibers
# with a relative FBC (RFBC) lower than the threshold :math:`\tau = 0.2`.
#
#
# Acknowledgments
# ---------------
# The techniques are developed in close collaboration with Pauly Ossenblok of
# the Academic Center of Epileptology Kempenhaeghe & Maastricht UMC+.
#
# References
# ----------
#
# .. footbibliography::
#



import warnings

import numpy as np
from numpy.testing import assert_array_almost_equal, assert_equal, assert_raises
import pytest

from dipy.core.gradients import generate_bvecs, gradient_table
from dipy.denoise import patch2self as p2s
from dipy.sims.voxel import multi_tensor
from dipy.testing import (
    assert_greater,
    assert_greater_equal,
    assert_less,
    assert_less_equal,
)
from dipy.testing.decorators import set_random_number_generator
from dipy.utils.optpkg import optional_package

sklearn, has_sklearn, _ = optional_package("sklearn")
needs_sklearn = pytest.mark.skipif(not has_sklearn, reason="Requires sklearn")


@needs_sklearn
@set_random_number_generator(1234)
def test_patch2self_random_noise(rng):
    S0 = 30 + 2 * rng.standard_normal((20, 20, 20, 50))

    bvals = np.repeat(30, 50)

    # shift = True
    for version in [1, 3]:
        extra_args = {"patch_radius": (0, 0, 0)} if version == 1 else {}
        S0den_shift = p2s.patch2self(
            S0,
            bvals,
            model="ols",
            shift_intensity=True,
            version=version,
            **extra_args,
        )

        assert_greater_equal(S0den_shift.min(), S0.min())
        assert_less_equal(np.round(S0den_shift.mean()), 30)

        # clip = True
        msg = "Both `clip_negative_vals` and `shift_intensity` .*"
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message=msg, category=UserWarning)
            S0den_clip = p2s.patch2self(
                S0,
                bvals,
                model="ols",
                clip_negative_vals=True,
                version=version,
                **extra_args,
            )

        assert_greater(S0den_clip.min(), S0.min())
        assert_equal(np.round(S0den_clip.mean()), 30)

        # both clip and shift = True, and int patch_radius
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message=msg, category=UserWarning)
            S0den_clip = p2s.patch2self(
                S0,
                bvals,
                model="ols",
                clip_negative_vals=True,
                shift_intensity=True,
                version=version,
                **extra_args,
            )

        assert_greater(S0den_clip.min(), S0.min())
        assert_equal(np.round(S0den_clip.mean()), 30)

        # both clip and shift = False
        S0den_clip = p2s.patch2self(
            S0,
            bvals,
            model="ols",
            clip_negative_vals=False,
            shift_intensity=False,
            version=version,
            **extra_args,
        )

        assert_greater(S0den_clip.min(), S0.min())
        assert_equal(np.round(S0den_clip.mean()), 30)


@needs_sklearn
@set_random_number_generator(1234)
def test_patch2self_boundary(rng):
    # patch2self preserves boundaries
    S0 = 100 + np.zeros((20, 20, 20, 20))
    noise = 2 * rng.standard_normal((20, 20, 20, 20))
    S0 += noise
    S0[:10, :10, :10, :10] = 300 + noise[:10, :10, :10, :10]

    bvals = np.repeat(100, 20)
    for version in [1, 3]:
        extra_args = {"patch_radius": (0, 0, 0)} if version == 1 else {}
        p2s.patch2self(S0, bvals, **extra_args)
        assert_greater(S0[9, 9, 9, 9], 290)
        assert_less(S0[10, 10, 10, 10], 110)


def rfiw_phantom(gtab, snr=None, rng=None):
    """rectangle fiber immersed in water"""
    # define voxel index
    slice_ind = np.zeros((10, 10, 8))
    slice_ind[4:7, 4:7, :] = 1
    slice_ind[4:7, 7, :] = 2
    slice_ind[7, 7, :] = 3
    slice_ind[7, 4:7, :] = 4
    slice_ind[7, 3, :] = 5
    slice_ind[4:7, 3, :] = 6
    slice_ind[3, 3, :] = 7
    slice_ind[3, 4:7, :] = 8
    slice_ind[3, 7, :] = 9

    # Define tissue diffusion parameters
    # Restricted diffusion
    ADr = 0.99e-3
    RDr = 0.0
    # Hindered diffusion
    ADh = 2.26e-3
    RDh = 0.87e-3
    # S0 value for tissue
    S1 = 50
    # Fraction between Restricted and Hindered diffusion
    fia = 0.51

    # Define water diffusion
    Dwater = 3e-3
    S2 = 100  # S0 value for water

    # Define tissue volume fraction for each voxel type (in index order)
    f = np.array([0.0, 1.0, 0.6, 0.18, 0.30, 0.15, 0.50, 0.35, 0.70, 0.42])

    # Define S0 for each voxel (in index order)
    S0 = S1 * f + S2 * (1 - f)

    # multi tensor simulations assume that each water pull as constant S0
    # since I am assuming that tissue and water voxels have different S0,
    # tissue volume fractions have to be adjusted to the measured f values when
    # constant S0 are assumed constant. Doing this correction, simulations will
    # be analogous to simulates that S0 are different for each media. (For more
    # details on this contact the phantom designer)
    f1 = f * S1 / S0

    mevals = np.array([[ADr, RDr, RDr], [ADh, RDh, RDh], [Dwater, Dwater, Dwater]])
    angles = [(0, 0, 1), (0, 0, 1), (0, 0, 1)]
    dwi = np.zeros(slice_ind.shape + (gtab.bvals.size,))
    for i in range(10):
        fractions = [
            f1[i] * fia * 100,
            f1[i] * (1 - fia) * 100,
            (1 - f1[i]) * 100,
        ]
        sig, direction = multi_tensor(
            gtab, mevals, S0=S0[i], angles=angles, fractions=fractions, snr=None
        )
        dwi[slice_ind == i, :] = sig
    if snr is None:
        return dwi
    else:
        sigma = S2 * 1.0 / snr
        n1 = rng.normal(0, sigma, size=dwi.shape)
        n2 = rng.normal(0, sigma, size=dwi.shape)
        return [
            np.sqrt((dwi / np.sqrt(2) + n1) ** 2 + (dwi / np.sqrt(2) + n2) ** 2),
            sigma,
        ]


@needs_sklearn
@set_random_number_generator(4321)
def test_phantom(rng):
    # generate a gradient table for phantom data
    directions8 = generate_bvecs(8)
    directions30 = generate_bvecs(20)
    directions60 = generate_bvecs(30)
    # Create full dataset parameters
    # (6 b-values = 0, 8 directions for b-value 300, 30 directions for b-value
    # 1000 and 60 directions for b-value 2000)
    bvals = np.hstack(
        (np.zeros(6), 300 * np.ones(8), 1000 * np.ones(20), 2000 * np.ones(30))
    )
    bvecs = np.vstack((np.zeros((6, 3)), directions8, directions30, directions60))
    gtab = gradient_table(bvals, bvecs=bvecs)

    dwi, sigma = rfiw_phantom(gtab, snr=10, rng=rng)
    dwi_den1 = p2s.patch2self(dwi, model="ridge", bvals=bvals, alpha=1.0, version=1)

    assert_less(np.max(dwi_den1) / sigma, np.max(dwi) / sigma)
    dwi_den2 = p2s.patch2self(dwi, model="ridge", bvals=bvals, alpha=0.7, version=1)

    assert_less(np.max(dwi_den2) / sigma, np.max(dwi) / sigma)
    assert_array_almost_equal(dwi_den1, dwi_den2, decimal=0)

    assert_raises(ValueError, p2s.patch2self, dwi, model="empty", bvals=bvals)

    # Try this with a sigma volume, instead of a scalar
    dwi_den = p2s.patch2self(dwi, bvals=bvals, model="ols", version=1)

    assert_less(np.max(dwi_den) / sigma, np.max(dwi) / sigma)


@needs_sklearn
def test_validate_patch_radius_and_version():
    data = np.random.rand(5, 5, 5, 10)
    bvals = np.zeros(10)

    test_cases = [
        {
            "patch_radius": 1,
            "version": 1,
            "tmp_dir": None,
            "expect_fail": False,
        },
        {
            "patch_radius": (1, 1, 1),
            "version": 1,
            "tmp_dir": None,
            "expect_fail": False,
        },
        {
            "patch_radius": (0, 0, 0),
            "version": 1,
            "tmp_dir": None,
            "expect_fail": False,
        },
        {
            "patch_radius": (0, 0, 0),
            "version": 3,
            "tmp_dir": None,
            "expect_fail": False,
        },
        {"patch_radius": 1, "version": 3, "tmp_dir": None, "expect_fail": True},
        {
            "patch_radius": (1, 1, 1),
            "version": 3,
            "tmp_dir": None,
            "expect_fail": True,
        },
        {
            "patch_radius": (0, 0, 0),
            "version": 3,
            "tmp_dir": "/nonexistent_dir",
            "expect_fail": True,
        },
        {
            "patch_radius": 1,
            "version": 1,
            "tmp_dir": "/some_temp_dir",
            "expect_fail": True,
        },
    ]

    for case in test_cases:
        patch_radius = case["patch_radius"]
        version = case["version"]
        tmp_dir = case["tmp_dir"]
        expect_fail = case["expect_fail"]

        if expect_fail:
            # Expecting a ValueError for this case
            with pytest.raises(ValueError):
                p2s.patch2self(
                    data,
                    bvals,
                    patch_radius=patch_radius,
                    version=version,
                    tmp_dir=tmp_dir,
                )
        else:
            # Expecting success
            try:
                result = p2s.patch2self(
                    data,
                    bvals,
                    patch_radius=patch_radius,
                    version=version,
                    tmp_dir=tmp_dir,
                )
                assert result.shape == data.shape, (
                    f"Shape mismatch with patch_radius={patch_radius}, "
                    f"version={version}, tmp_dir={tmp_dir}"
                )
            except ValueError:
                pytest.fail(
                    f"Unexpected ValueError with patch_radius={patch_radius}, "
                    f"version={version}, tmp_dir={tmp_dir}"
                )


@needs_sklearn
def test_single_slice_data():
    for version in [1, 3]:
        # Create single-slice 4D data with shape (64, 64, 1, 10)
        single_slice_data = np.random.rand(64, 64, 1, 10).astype(np.float32)
        bvals = np.array([0] * 5 + [1000] * 5)  # Simulate bvals for testing

        # Run the Patch2Self function
        denoised_data = p2s.patch2self(single_slice_data, bvals, version=version)

        assert denoised_data.shape == single_slice_data.shape, (
            f"Expected shape {single_slice_data.shape} for version {version}, "
            f"but got {denoised_data.shape}."
        )



"""Benchmarks for functions related to streamline in ``dipy.tracking``module."""

import numpy as np

from dipy.data import get_fnames
from dipy.io.streamline import load_tractogram
from dipy.tracking import Streamlines
from dipy.tracking.streamline import length, set_number_of_points
from dipy.tracking.streamlinespeed import compress_streamlines


class BenchStreamlines:
    def setup(self):
        rng = np.random.default_rng(42)
        nb_streamlines = 20000
        min_nb_points = 2
        max_nb_points = 100

        def generate_streamlines(nb_streamlines, min_nb_points, max_nb_points, rng):
            streamlines = [
                rng.random((rng.integers(min_nb_points, max_nb_points), 3))
                for _ in range(nb_streamlines)
            ]
            return streamlines

        self.data = {}
        self.data["rng"] = rng
        self.data["nb_streamlines"] = nb_streamlines
        self.data["streamlines"] = generate_streamlines(
            nb_streamlines, min_nb_points, max_nb_points, rng=rng
        )
        self.data["streamlines_arrseq"] = Streamlines(self.data["streamlines"])

        fname = get_fnames(name="fornix")
        fornix = load_tractogram(fname, "same", bbox_valid_check=False).streamlines

        self.fornix_streamlines = Streamlines(fornix)

    def time_set_number_of_points(self):
        streamlines = self.data["streamlines"]
        set_number_of_points(streamlines, nb_points=50)

    def time_set_number_of_points_arrseq(self):
        streamlines = self.data["streamlines_arrseq"]
        set_number_of_points(streamlines, nb_points=50)

    def time_length(self):
        streamlines = self.data["streamlines"]
        length(streamlines)

    def time_length_arrseq(self):
        streamlines = self.data["streamlines_arrseq"]
        length(streamlines)

    def time_compress_streamlines(self):
        compress_streamlines(self.fornix_streamlines)



import logging
import sys

import numpy as np

from dipy.core.gradients import extract_b0, gradient_table
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, save_nifti
from dipy.nn.deepn4 import DeepN4
from dipy.nn.evac import EVACPlus
from dipy.workflows.workflow import Workflow


class EVACPlusFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "evacplus"

    def run(
        self,
        input_files,
        save_masked=False,
        out_dir="",
        out_mask="brain_mask.nii.gz",
        out_masked="dwi_masked.nii.gz",
    ):
        """Extract brain using EVAC+.

        See :footcite:p:`Park2024` for further details about EVAC+.

        Parameters
        ----------
        input_files : string
            Path to the input volumes. This path may contain wildcards to
            process multiple inputs at once.
        save_masked : bool, optional
            Save mask.
        out_dir : string, optional
            Output directory.
        out_mask : string, optional
            Name of the mask volume to be saved.
        out_masked : string, optional
            Name of the masked volume to be saved.

        References
        ----------
        .. footbibliography::
        """
        io_it = self.get_io_iterator()
        empty_flag = True

        for fpath, mask_out_path, masked_out_path in io_it:
            logging.info(f"Applying evac+ brain extraction on {fpath}")

            data, affine, img, voxsize = load_nifti(
                fpath, return_img=True, return_voxsize=True
            )
            evac = EVACPlus()
            mask_volume = evac.predict(data, affine, voxsize=voxsize)
            masked_volume = mask_volume * data

            save_nifti(mask_out_path, mask_volume.astype(np.float64), affine)

            logging.info(f"Mask saved as {mask_out_path}")

            if save_masked:
                save_nifti(masked_out_path, masked_volume, affine, hdr=img.header)

                logging.info(f"Masked volume saved as {masked_out_path}")
            empty_flag = False
        if empty_flag:
            raise ValueError(
                "All output paths exists."
                " If you want to overwrite "
                "please use the --force option."
            )

        return io_it


class BiasFieldCorrectionFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "bias_field_correction"

    def run(
        self,
        input_files,
        bval=None,
        bvec=None,
        method="n4",
        threshold=0.5,
        use_cuda=False,
        verbose=False,
        out_dir="",
        out_corrected="biasfield_corrected.nii.gz",
    ):
        """Correct bias field.

        Parameters
        ----------
        input_files : string
            Path to the input volumes. This path may contain wildcards to
            process multiple inputs at once.
        bval : string, optional
            Path to the b-value file.
        bvec : string, optional
            Path to the b-vector file.
        method : string, optional
            Bias field correction method. Choose from:
                - 'n4': DeepN4 bias field correction.
                  See :footcite:p:`Kanakaraj2024` for more details.
                - 'b0': B0 bias field correction via normalization.

            'n4' method is recommended for T1-weighted images where 'b0' method
            is recommended for diffusion-weighted images.
        threshold : float, optional
            Threshold for cleaning the final correction field in DeepN4 method.
        use_cuda : bool, optional
            Use CUDA for DeepN4 bias field correction.
        verbose : bool, optional
            Print verbose output.
        out_dir : string, optional
            Output directory.
        out_corrected : string, optional
            Name of the corrected volume to be saved.

        References
        ----------
        .. footbibliography::

        """
        io_it = self.get_io_iterator()

        if method.lower() not in ["n4", "b0"]:
            logging.error("Unknown bias field correction method. Choose from 'n4, b0'.")
            sys.exit(1)

        prefix = "t1" if method.lower() == "n4" else "dwi"
        for i, name in enumerate(self.flat_outputs):
            if name.endswith("biasfield_corrected.nii.gz"):
                self.flat_outputs[i] = name.replace(
                    "biasfield_corrected.nii.gz", f"{prefix}_biasfield_corrected.nii.gz"
                )

        self.update_flat_outputs(self.flat_outputs, io_it)
        for fpath, corrected_out_path in io_it:
            logging.info(f"Applying bias field correction on {fpath}")

            data, affine, img, voxsize = load_nifti(
                fpath, return_img=True, return_voxsize=True
            )

            corrected_data = None
            if method.lower() == "b0":
                bvals, bvecs = read_bvals_bvecs(bval, bvec)
                gtab = gradient_table(bvals, bvecs=bvecs)
                b0 = extract_b0(data, gtab.b0s_mask)
                for i in range(data.shape[-1]):
                    data[..., i] = np.divide(
                        data[..., i],
                        b0,
                        out=np.zeros_like(data[..., i]).astype(float),
                        where=b0 != 0,
                    )
                corrected_data = data
            elif method.lower() == "n4":
                deepn4_model = DeepN4(verbose=verbose, use_cuda=use_cuda)
                deepn4_model.fetch_default_weights()
                corrected_data = deepn4_model.predict(
                    data, affine, voxsize=voxsize, threshold=threshold
                )

            save_nifti(corrected_out_path, corrected_data, affine, hdr=img.header)
            logging.info(f"Corrected volume saved as {corrected_out_path}")

        return io_it



from dipy.direction.closest_peak_direction_getter cimport PmfGenDirectionGetter


cdef class ProbabilisticDirectionGetter(PmfGenDirectionGetter):
    cdef:
        double[:, :] vertices


cdef class DeterministicMaximumDirectionGetter(ProbabilisticDirectionGetter):
    pass



from dipy.core.onetime import auto_attr
from dipy.testing.decorators import warning_for_keywords


class Cache:
    """Cache values based on a key object (such as a sphere or gradient table).

    Notes
    -----
    This class is meant to be used as a mix-in::

        class MyModel(Model, Cache):
            pass

        class MyModelFit(Fit):
            pass

    Inside a method on the fit, typical usage would be::

        def odf(sphere):
            M = self.model.cache_get('odf_basis_matrix', key=sphere)

            if M is None:
                M = self._compute_basis_matrix(sphere)
                self.model.cache_set('odf_basis_matrix', key=sphere, value=M)

    """

    # We use this method instead of __init__ to construct the cache, so
    # that the class can be used as a mixin, without having to worry about
    # calling the super-class constructor
    @auto_attr
    def _cache(self):
        return {}

    def cache_set(self, tag, key, value):
        """Store a value in the cache.

        Parameters
        ----------
        tag : str
            Description of the cached value.
        key : object
            Key object used to look up the cached value.
        value : object
            Value stored in the cache for each unique combination
            of ``(tag, key)``.

        Examples
        --------
        >>> def compute_expensive_matrix(parameters):
        ...     # Imagine the following computation is very expensive
        ...     return (p**2 for p in parameters)

        >>> c = Cache()

        >>> parameters = (1, 2, 3)
        >>> X1 = compute_expensive_matrix(parameters)

        >>> c.cache_set('expensive_matrix', parameters, X1)
        >>> X2 = c.cache_get('expensive_matrix', parameters)

        >>> X1 is X2
        True

        """
        self._cache[(tag, key)] = value

    @warning_for_keywords()
    def cache_get(self, tag, key, *, default=None):
        """Retrieve a value from the cache.

        Parameters
        ----------
        tag : str
            Description of the cached value.
        key : object
            Key object used to look up the cached value.
        default : object
            Value to be returned if no cached entry is found.

        Returns
        -------
        v : object
            Value from the cache associated with ``(tag, key)``.  Returns
            `default` if no cached entry is found.

        """
        return self._cache.get((tag, key), default)

    def cache_clear(self):
        """Clear the cache."""
        self._cache = {}



# cython: boundscheck=False
# cython: initializedcheck=False
# cython: wraparound=False
# cython: Nonecheck=False

from libc.stdio cimport printf

cimport ctime
cimport cython
from cython.parallel import prange
import numpy as np
cimport numpy as cnp

from dipy.direction.pmf cimport PmfGen
from dipy.tracking.stopping_criterion cimport StoppingCriterion
from dipy.utils cimport fast_numpy

from dipy.tracking.stopping_criterion cimport (StreamlineStatus,
                                               StoppingCriterion,
                                               TRACKPOINT,
                                               ENDPOINT,
                                               OUTSIDEIMAGE,
                                               INVALIDPOINT,
                                               VALIDSTREAMLIME,
                                               INVALIDSTREAMLIME)
from dipy.tracking.tracker_parameters cimport (TrackerParameters,
                                               func_ptr,
                                               SUCCESS,
                                               FAIL)

from nibabel.streamlines import ArraySequence as Streamlines

from libc.stdlib cimport malloc, free
from libc.string cimport memcpy, memset
from libc.math cimport ceil
from libc.stdio cimport printf


def generate_tractogram(double[:,::1] seed_positions,
                        double[:,::1] seed_directions,
                        StoppingCriterion sc,
                        TrackerParameters params,
                        PmfGen pmf_gen,
                        affine,
                        int nbr_threads=0,
                        float buffer_frac=1.0,
                        bint save_seeds=0):
    """Generate a tractogram from a set of seed points and directions.

    Parameters
    ----------
    seed_positions : ndarray
        Seed positions for the streamlines.
    seed_directions : ndarray
        Seed directions for the streamlines.
    sc : StoppingCriterion
        Stopping criterion for the streamlines.
    params : TrackerParameters
        Parameters for the streamline generation.
    pmf_gen : PmfGen
        Probability mass function generator.
    affine : ndarray
        Affine transformation for the streamlines.
    nbr_threads : int, optional
        Number of threads to use for streamline generation.
    buffer_frac : float, optional
        Fraction of the seed points to process in each iteration.
    save_seeds : bool, optional
        If True, return seeds alongside streamlines

    Yields
    ------
    streamlines : Streamlines
        Streamlines generated from the seed points.
    seeds : ndarray, optional
        seed points associated with the generated streamlines.

    """
    cdef:
        cnp.npy_intp _len = seed_positions.shape[0]
        cnp.npy_intp _plen = int(ceil(_len * buffer_frac))
        cnp.npy_intp i, seed_start, seed_end
        double** streamlines_arr
        int* length_arr
        StreamlineStatus* status_arr

    if buffer_frac <=0 or buffer_frac > 1:
        raise ValueError("buffer_frac must > 0 and <= 1.")

    lin_T = affine[:3, :3].T.copy()
    offset = affine[:3, 3].copy()

    inv_affine = np.linalg.inv(affine)
    seed_positions = np.dot(seed_positions, inv_affine[:3, :3].T.copy())
    seed_positions += inv_affine[:3, 3]

    seed_start = 0
    seed_end = _plen
    while seed_start < _len:
        streamlines_arr = <double**> malloc(_plen * sizeof(double*))
        length_arr = <int*> malloc(_plen * sizeof(int))
        status_arr = <StreamlineStatus*> malloc(_plen * sizeof(int))

        if streamlines_arr == NULL or length_arr == NULL or status_arr == NULL:
            raise MemoryError("Memory allocation failed")

        generate_tractogram_c(seed_positions[seed_start:seed_end],
                              seed_directions[seed_start:seed_end],
                              nbr_threads, sc, params, pmf_gen,
                              streamlines_arr, length_arr, status_arr)

        for i in range(seed_end - seed_start):
            if ((status_arr[i] == VALIDSTREAMLIME or params.return_all)
                and (length_arr[i] >= params.min_len
                     and length_arr[i] <= params.max_len)):
                s = np.asarray(<cnp.float_t[:length_arr[i]*3]> streamlines_arr[i])
                track = s.copy().reshape((-1,3))
                if save_seeds:
                    yield np.dot(track, lin_T) + offset, np.dot(seed_positions[seed_start + i], lin_T) + offset
                else:
                    yield np.dot(track, lin_T) + offset
            free(streamlines_arr[i])

        free(streamlines_arr)
        free(length_arr)
        free(status_arr)

        seed_start += _plen
        seed_end += _plen
        if seed_end > _len:
            seed_end = _len


cdef void generate_tractogram_c(double[:,::1] seed_positions,
                                double[:,::1] seed_directions,
                               int nbr_threads,
                               StoppingCriterion sc,
                               TrackerParameters params,
                               PmfGen pmf_gen,
                               double** streamlines,
                               int* lengths,
                               StreamlineStatus* status):
    """Generate a tractogram from a set of seed points and directions.

    This is the C implementation of the generate_tractogram function.

    Parameters
    ----------
    seed_positions : ndarray
        Seed positions for the streamlines.
    seed_directions : ndarray
        Seed directions for the streamlines.
    nbr_threads : int
        Number of threads to use for streamline generation.
    sc : StoppingCriterion
        Stopping criterion for the streamlines.
    params : TrackerParameters
        Parameters for the streamline generation.
    pmf_gen : PmfGen
        Probability mass function generator.
    streamlines : list
        List to store the generated streamlines.
    lengths : list
        List to store the lengths of the generated streamlines.
    status : list
        List to store the status of the generated streamlines.

    """
    cdef:
        cnp.npy_intp _len=seed_positions.shape[0]
        cnp.npy_intp i

    if nbr_threads<= 0:
        nbr_threads = 0
    for i in prange(_len, nogil=True, num_threads=nbr_threads):
        stream = <double*> malloc((params.max_len * 3 * 2 + 1) * sizeof(double))
        stream_idx = <int*> malloc(2 * sizeof(int))
        status[i] = generate_local_streamline(&seed_positions[i][0],
                                              &seed_directions[i][0],
                                              stream,
                                              stream_idx,
                                              sc,
                                              params,
                                              pmf_gen)

        # copy the streamlines points from the buffer to a 1d vector of the streamline length
        lengths[i] = stream_idx[1] - stream_idx[0] + 1
        streamlines[i] = <double*> malloc(lengths[i] * 3 * sizeof(double))
        memcpy(&streamlines[i][0], &stream[stream_idx[0] * 3], lengths[i] * 3 * sizeof(double))
        free(stream)
        free(stream_idx)


cdef StreamlineStatus generate_local_streamline(double* seed,
                                                double* direction,
                                                double* stream,
                                                int* stream_idx,
                                                StoppingCriterion sc,
                                                TrackerParameters params,
                                                PmfGen pmf_gen) noexcept nogil:
    """Generate a unique streamline from a seed point and direction.

    This is the C implementation.

    Parameters
    ----------
    seed : ndarray
        Seed point for the streamline.
    direction : ndarray
        Seed direction for the streamline.
    stream : ndarray
        Buffer to store the generated streamline.
    stream_idx : ndarray
        Buffer to store the indices of the generated streamline.
    sc : StoppingCriterion
        Stopping criterion for the streamline.
    params : TrackerParameters
        Parameters for the streamline generation.
    pmf_gen : PmfGen
        Probability mass function generator.

    """
    cdef:
        cnp.npy_intp i, j
        cnp.npy_uint32 s_random_seed
        double[3] point
        double[3] voxdir
        double voxdir_norm
        double* stream_data
        StreamlineStatus status_forward, status_backward
        fast_numpy.RNGState rng

    # set the random generator
    if params.random_seed > 0:
        s_random_seed = int(
            (seed[0] * 2 + seed[1] * 3 + seed[2] * 5) * params.random_seed
            )
    else:
        s_random_seed = <cnp.npy_uint32>ctime.time_ns()

    fast_numpy.seed_rng(&rng, s_random_seed)

    # set the initial position
    fast_numpy.copy_point(seed, point)
    fast_numpy.copy_point(direction, voxdir)
    fast_numpy.copy_point(seed, &stream[params.max_len * 3])
    stream_idx[0] = stream_idx[1] = params.max_len

    # the input direction is invalid
    voxdir_norm = fast_numpy.norm(voxdir)
    if voxdir_norm < 0.99 or voxdir_norm > 1.01:
        return INVALIDSTREAMLIME

    # forward tracking
    stream_data = <double*> malloc(100 * sizeof(double))
    memset(stream_data, 0, 100 * sizeof(double))
    status_forward = TRACKPOINT
    for i in range(1, params.max_len):
        if params.tracker(&point[0], &voxdir[0], params, stream_data, pmf_gen, &rng) == FAIL:
            break
        # update position
        for j in range(3):
            point[j] += voxdir[j] * params.inv_voxel_size[j] * params.step_size
        fast_numpy.copy_point(point, &stream[(params.max_len + i )* 3])

        status_forward = sc.check_point_c(point, &rng)
        if (status_forward == ENDPOINT or
            status_forward == INVALIDPOINT or
            status_forward == OUTSIDEIMAGE):
            break
    stream_idx[1] = params.max_len + i - 1
    free(stream_data)

    # backward tracking
    stream_data = <double*> malloc(100 * sizeof(double))
    memset(stream_data, 0, 100 * sizeof(double))

    fast_numpy.copy_point(seed, point)
    fast_numpy.copy_point(direction, voxdir)
    if i > 1:
        # Use the first selected orientation for the backward tracking segment
        for j in range(3):
            voxdir[j] = stream[(params.max_len + 1) * 3 + j] - stream[params.max_len * 3 + j]
        fast_numpy.normalize(voxdir)

    # flip the initial direction for backward streamline segment
    for j in range(3):
        voxdir[j] = voxdir[j] * -1

    status_backward = TRACKPOINT
    for i in range(1, params.max_len):
        if params.tracker(&point[0], &voxdir[0], params, stream_data, pmf_gen, &rng) == FAIL:
            break
        # update position
        for j in range(3):
            point[j] += voxdir[j] * params.inv_voxel_size[j] * params.step_size
        fast_numpy.copy_point(point, &stream[(params.max_len - i )* 3])

        status_backward = sc.check_point_c(point, &rng)
        if (status_backward == ENDPOINT or
            status_backward == INVALIDPOINT or
            status_backward == OUTSIDEIMAGE):
            break
    stream_idx[0] = params.max_len - i + 1
    free(stream_data)

    # check for valid streamline ending status
    if ((status_backward == ENDPOINT or status_backward == OUTSIDEIMAGE)
        and (status_forward == ENDPOINT or status_forward == OUTSIDEIMAGE)):
        return VALIDSTREAMLIME
    return INVALIDSTREAMLIME


cdef void prepare_pmf(double* pmf,
                      double* point,
                      PmfGen pmf_gen,
                      double pmf_threshold,
                      int pmf_len) noexcept nogil:
    """Prepare the probability mass function for streamline generation.

    Parameters
    ----------
    pmf : ndarray
        Probability mass function.
    point : ndarray
        Current tracking position.
    pmf_gen : PmfGen
        Probability mass function generator.
    pmf_threshold : float
        Threshold for the probability mass function.
    pmf_len : int
        Length of the probability mass function.

    """
    cdef:
        cnp.npy_intp i
        double absolute_pmf_threshold
        double max_pmf=0

    pmf = pmf_gen.get_pmf_c(point, pmf)

    for i in range(pmf_len):
        if pmf[i] > max_pmf:
            max_pmf = pmf[i]
    absolute_pmf_threshold = pmf_threshold * max_pmf

    for i in range(pmf_len):
        if pmf[i] < absolute_pmf_threshold:
            pmf[i] = 0.0



"""
=====================================================
Using the RESTORE algorithm for robust tensor fitting
=====================================================

The diffusion tensor model takes into account certain kinds of noise (thermal),
but not other kinds, such as "physiological" noise. For example, if a subject
moves during acquisition of one of the diffusion-weighted samples, this
might have a substantial effect on the parameters of the tensor fit calculated
in all voxels in the brain for that subject. One of the pernicious consequences
of this is that it can lead to wrong interpretation of group differences. For
example, some groups of participants (e.g. young children, patient groups,
etc.) are particularly prone to motion and differences in tensor parameters and
derived statistics (such as FA) due to motion would be confounded with actual
differences in the physical properties of the white matter. An example of this
was shown in a paper by :footcite:t:`Yendiki2014`.

One of the strategies to deal with this problem is to apply an automatic method
for detecting outliers in the data, excluding these outliers and refitting the
model without the presence of these outliers. This is often referred to as
"robust model fitting". One of the common algorithms for robust tensor fitting
is called RESTORE, and was first proposed by :footcite:p:`Chang2005`.

In the following example, we will demonstrate how to use RESTORE on a simulated
dataset, which we will corrupt by adding intermittent noise.

We start by importing a few of the libraries we will use.

- The module ``dipy.reconst.dti`` contains the implementation of tensor
  fitting, including an implementation of the RESTORE algorithm.
- The module ``dipy.data`` is used for small datasets that we use in tests and
  examples.
- ``dipy.io.image`` is for loading / saving imaging datasets
- ``dipy.io.gradients`` is for loading / saving our bvals and bvecs
- ``dipy.viz`` package is used for 3D visualization and matplotlib for 2D
  visualizations:
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.core.gradients import gradient_table
import dipy.data as dpd
import dipy.denoise.noise_estimate as ne
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
import dipy.reconst.dti as dti
from dipy.viz import actor, window

# Enables/disables interactive visualization
interactive = False

###############################################################################
# If needed, the ``fetch_stanford_hardi`` function will download the raw dMRI
# dataset of a single subject. The size of this dataset is 87 MBytes. You only
# need to fetch once.

hardi_fname, hardi_bval_fname, hardi_bvec_fname = dpd.get_fnames(name="stanford_hardi")
data, affine = load_nifti(hardi_fname)

bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# We initialize a DTI model class instance using the gradient table used in
# the measurement. By default, ``dti.TensorModel`` will use a weighted
# least-squares algorithm (described in :footcite:p:`Chang2005`) to fit the
# parameters of the model. We initialize this model as a baseline for
# comparison of noise-corrupted models:

dti_wls = dti.TensorModel(gtab)

###############################################################################
# For the purpose of this example, we will focus on the data from a region of
# interest (ROI) surrounding the Corpus Callosum. We define that ROI as the
# following indices:

roi_idx = (slice(20, 50), slice(55, 85), slice(38, 39))

###############################################################################
# And use them to index into the data:

data = data[roi_idx]

###############################################################################
# This dataset is not very noisy, so we will artificially corrupt it to
# simulate the effects of "physiological" noise, such as subject motion. But
# first, let's establish a baseline, using the data as it is:

fit_wls = dti_wls.fit(data)

fa1 = fit_wls.fa
evals1 = fit_wls.evals
evecs1 = fit_wls.evecs
cfa1 = dti.color_fa(fa1, evecs1)
sphere = dpd.default_sphere

###############################################################################
# We visualize the ODFs in the ROI using ``dipy.viz`` module:

scene = window.Scene()
scene.add(
    actor.tensor_slicer(evals1, evecs1, scalar_colors=cfa1, sphere=sphere, scale=0.3)
)
window.record(scene=scene, out_path="tensor_ellipsoids_wls.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Tensor Ellipsoids.

scene.clear()

###############################################################################
# Next, we corrupt the data with some noise. To simulate a subject that moves
# intermittently, we will replace a few of the images with a very low signal

noisy_data = np.copy(data)
noisy_idx = slice(-10, None)  # The last 10 volumes are corrupted
noisy_data[..., noisy_idx] = 1.0

###############################################################################
# We use the same model to fit this noisy data

fit_wls_noisy = dti_wls.fit(noisy_data)
fa2 = fit_wls_noisy.fa
evals2 = fit_wls_noisy.evals
evecs2 = fit_wls_noisy.evecs
cfa2 = dti.color_fa(fa2, evecs2)

scene = window.Scene()
scene.add(
    actor.tensor_slicer(evals2, evecs2, scalar_colors=cfa2, sphere=sphere, scale=0.3)
)
window.record(scene=scene, out_path="tensor_ellipsoids_wls_noisy.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# In places where the tensor model is particularly sensitive to noise, the
# resulting tensor field will be distorted
#
# .. rst-class:: centered small fst-italic fw-semibold
#
# Tensor Ellipsoids from noisy data.
#
#
# To estimate the parameters from the noisy data using RESTORE, we need to
# estimate what would be a reasonable amount of noise to expect in the
# measurement. To do that, we use the ``dipy.denoise.noise_estimate`` module:

sigma = ne.estimate_sigma(data)

###############################################################################
# This estimate of the standard deviation will be used by the RESTORE
# algorithm to identify the outliers in each voxel and is given as an input
# when initializing the TensorModel object:

dti_restore = dti.TensorModel(gtab, fit_method="RESTORE", sigma=sigma)
fit_restore_noisy = dti_restore.fit(noisy_data)
fa3 = fit_restore_noisy.fa
evals3 = fit_restore_noisy.evals
evecs3 = fit_restore_noisy.evecs
cfa3 = dti.color_fa(fa3, evecs3)

scene = window.Scene()
scene.add(
    actor.tensor_slicer(evals3, evecs3, scalar_colors=cfa3, sphere=sphere, scale=0.3)
)
print("Saving illustration as tensor_ellipsoids_restore_noisy.png")
window.record(
    scene=scene, out_path="tensor_ellipsoids_restore_noisy.png", size=(600, 600)
)
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold

# Tensor Ellipsoids from noisy data recovered with RESTORE.


# The tensor field looks rather restored to its noiseless state in this
# image, but to convince ourselves further that this did the right thing, we
# will compare  the distribution of FA in this region relative to the
# baseline, using the RESTORE estimate and the WLS estimate
# :footcite:p:`Chung2006`.

fig_hist, ax = plt.subplots(1)
ax.hist(np.ravel(fa2), color="b", histtype="step", label="WLS")
ax.hist(np.ravel(fa3), color="r", histtype="step", label="RESTORE")
ax.hist(np.ravel(fa1), color="g", histtype="step", label="Original")
ax.set_xlabel("Fractional Anisotropy")
ax.set_ylabel("Count")
plt.legend()
fig_hist.savefig("dti_fa_distributions.png")

###############################################################################
# This demonstrates that RESTORE can recover a distribution of FA that more
# closely resembles the baseline distribution of the noiseless signal, and
# demonstrates the utility of the method to data with intermittent
# noise. Importantly, this method assumes that the tensor is a good
# representation of the diffusion signal in the data. If you have reason to
# believe this is not the case (for example, you have data with very high b
# values and you are particularly interested in locations in the brain in which
# fibers cross), you might want to use a different method to fit your data.
#
#
# References
# ----------
#
# .. footbibliography::
#



"""
==========================
DKI MultiTensor Simulation
==========================

In this example we show how to simulate the Diffusion Kurtosis Imaging (DKI)
data of a single voxel. DKI captures information about the non-Gaussian
properties of water diffusion which is a consequence of the existence of tissue
barriers and compartments. In these simulations compartmental heterogeneity is
taken into account by modeling different compartments for the intra- and
extra-cellular media of two populations of fibers. These simulations are
performed according to :footcite:p:`NetoHenriques2015`.

We first import all relevant modules.
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.reconst.dti import decompose_tensor, from_lower_triangular
from dipy.sims.voxel import multi_tensor_dki, single_tensor

###############################################################################
# For the simulation we will need a GradientTable with the b-values and
# b-vectors. Here we use the GradientTable of the sample DIPY_ dataset
# ``small_64D``.

fimg, fbvals, fbvecs = get_fnames(name="small_64D")
bvals, bvecs = read_bvals_bvecs(fbvals, fbvecs)

###############################################################################
# DKI requires data from more than one non-zero b-value. Since the dataset
# ``small_64D`` was acquired with one non-zero b-value we artificially produce
# a second non-zero b-value.

bvals = np.concatenate((bvals, bvals * 2), axis=0)
bvecs = np.concatenate((bvecs, bvecs), axis=0)

###############################################################################
# The b-values and gradient directions are then converted to DIPY's
# ``GradientTable`` format.

gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# In ``mevals`` we save the eigenvalues of each tensor. To simulate crossing
# fibers with two different media (representing intra and extra-cellular
# media), a total of four components have to be taken in to account (i.e. the
# first two compartments correspond to the intra and extra cellular media for
# the first fiber population while the others correspond to the media of the
# second fiber population)

mevals = np.array(
    [
        [0.00099, 0, 0],
        [0.00226, 0.00087, 0.00087],
        [0.00099, 0, 0],
        [0.00226, 0.00087, 0.00087],
    ]
)

###############################################################################
# In ``angles`` we save in polar coordinates (:math:`\theta, \phi`) the
# principal axis of each compartment tensor. To simulate crossing fibers at
# 70$^{\circ}$ the compartments of the first fiber are aligned to the X-axis
# while the compartments of the second fiber are aligned to the X-Z plane with
# an angular deviation of 70$^{\\circ}$ from the first one.

angles = [(90, 0), (90, 0), (20, 0), (20, 0)]

###############################################################################
# In ``fractions`` we save the percentage of the contribution of each
# compartment, which is computed by multiplying the percentage of contribution
# of each fiber population and the water fraction of each different medium

fie = 0.49  # intra-axonal water fraction
fractions = [fie * 50, (1 - fie) * 50, fie * 50, (1 - fie) * 50]

###############################################################################
# Having defined the parameters for all tissue compartments, the elements of
# the diffusion tensor (DT), the elements of the kurtosis tensor (KT) and the
# DW signals simulated from the DKI model can be obtain using the function
# ``multi_tensor_dki``.

signal_dki, dt, kt = multi_tensor_dki(
    gtab, mevals, S0=200, angles=angles, fractions=fractions, snr=None
)

###############################################################################
# We can also add Rician noise with a specific SNR.

signal_noisy, dt, kt = multi_tensor_dki(
    gtab, mevals, S0=200, angles=angles, fractions=fractions, snr=10
)

###############################################################################
# For comparison purposes, we also compute the DW signal if only the diffusion
# tensor components are taken into account. For this we use DIPY's function
# ``single_tensor`` which requires that dt is decomposed into its eigenvalues
# and eigenvectors.

dt_evals, dt_evecs = decompose_tensor(from_lower_triangular(dt))
signal_dti = single_tensor(gtab, S0=200, evals=dt_evals, evecs=dt_evecs, snr=None)

###############################################################################
# Finally, we can visualize the values of the different version of simulated
# signals for all assumed gradient directions and bvalues.

plt.plot(signal_dti, label="noiseless dti")
plt.plot(signal_dki, label="noiseless dki")
plt.plot(signal_noisy, label="with noise")
plt.legend()
plt.savefig("simulated_dki_signal.png", bbox_inches="tight")


###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Simulated signals obtain from the DTI and DKI models.
#
#
# Non-Gaussian diffusion properties in tissues are responsible to smaller
# signal attenuation for larger bvalues when compared to signal attenuation
# from free Gaussian water diffusion. This feature can be shown from the
# figure above, since signals simulated from the DKI models reveals larger DW
# signal intensities than the signals obtained only from the diffusion tensor
# components.
#
# References
# ----------
#
# .. footbibliography::
#



from nibabel.affines import apply_affine, from_matvec
import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_array_almost_equal,
    assert_array_equal,
    assert_equal,
    assert_raises,
)
from scipy.ndimage import map_coordinates

from dipy.align import floating, imwarp, vector_fields as vfu
from dipy.align.parzenhist import sample_domain_regular
from dipy.align.transforms import regtransforms
from dipy.core import geometry
from dipy.testing.decorators import set_random_number_generator


@set_random_number_generator(3921116)
def test_random_displacement_field_2d(rng):
    from_shape = (25, 32)
    to_shape = (33, 29)

    # Create grid coordinates
    x_0 = np.asarray(range(from_shape[0]))
    x_1 = np.asarray(range(from_shape[1]))
    X = np.empty((3,) + from_shape, dtype=np.float64)
    _O = np.ones(from_shape)
    X[0, ...] = x_0[:, None] * _O
    X[1, ...] = x_1[None, :] * _O
    X[2, ...] = 1

    # Create an arbitrary image-to-space transform
    t = 0.15  # translation factor

    trans = np.array(
        [[1, 0, -t * from_shape[0]], [0, 1, -t * from_shape[1]], [0, 0, 1]]
    )
    trans_inv = np.linalg.inv(trans)

    for theta in [-1 * np.pi / 6.0, 0.0, np.pi / 5.0]:  # rotation angle
        for s in [0.83, 1.3, 2.07]:  # scale
            ct = np.cos(theta)
            st = np.sin(theta)

            rot = np.array([[ct, -st, 0], [st, ct, 0], [0, 0, 1]])

            scale = np.array([[1 * s, 0, 0], [0, 1 * s, 0], [0, 0, 1]])

            from_grid2world = trans_inv.dot(scale.dot(rot.dot(trans)))
            to_grid2world = from_grid2world.dot(scale)
            to_world2grid = np.linalg.inv(to_grid2world)

            field, assignment = vfu.create_random_displacement_2d(
                np.array(from_shape, dtype=np.int32),
                from_grid2world,
                np.array(to_shape, dtype=np.int32),
                to_grid2world,
                rng=rng,
            )
            field = np.array(field, dtype=floating)
            assignment = np.array(assignment)
            # Verify the assignments are inside the requested region
            assert_equal(0, (assignment < 0).sum())
            for i in range(2):
                assert_equal(0, (assignment[..., i] >= to_shape[i]).sum())

            # Compute the warping coordinates (see warp_2d documentation)
            Y = np.apply_along_axis(from_grid2world.dot, 0, X)[0:2, ...]
            Z = np.zeros_like(X)
            Z[0, ...] = Y[0, ...] + field[..., 0]
            Z[1, ...] = Y[1, ...] + field[..., 1]
            Z[2, ...] = 1
            W = np.apply_along_axis(to_world2grid.dot, 0, Z)[0:2, ...]

            # Verify the claimed assignments are correct
            assert_array_almost_equal(W[0, ...], assignment[..., 0], 5)
            assert_array_almost_equal(W[1, ...], assignment[..., 1], 5)

    # Test exception is raised when the affine transform matrix is not valid
    valid = np.zeros((2, 3), dtype=np.float64)
    invalid = np.zeros((2, 2), dtype=np.float64)
    shape = np.array(from_shape, dtype=np.int32)
    assert_raises(
        ValueError,
        vfu.create_random_displacement_2d,
        shape,
        invalid,
        shape,
        valid,
        rng=rng,
    )
    assert_raises(
        ValueError,
        vfu.create_random_displacement_2d,
        shape,
        valid,
        shape,
        invalid,
        rng=rng,
    )


@set_random_number_generator(7127562)
def test_random_displacement_field_3d(rng):
    from_shape = (25, 32, 31)
    to_shape = (33, 29, 35)

    # Create grid coordinates
    x_0 = np.asarray(range(from_shape[0]))
    x_1 = np.asarray(range(from_shape[1]))
    x_2 = np.asarray(range(from_shape[2]))
    X = np.empty((4,) + from_shape, dtype=np.float64)
    _O = np.ones(from_shape)
    X[0, ...] = x_0[:, None, None] * _O
    X[1, ...] = x_1[None, :, None] * _O
    X[2, ...] = x_2[None, None, :] * _O
    X[3, ...] = 1

    # Select an arbitrary rotation axis
    axis = np.array([0.5, 2.0, 1.5])

    # Create an arbitrary image-to-space transform
    t = 0.15  # translation factor

    trans = np.array(
        [
            [1, 0, 0, -t * from_shape[0]],
            [0, 1, 0, -t * from_shape[1]],
            [0, 0, 1, -t * from_shape[2]],
            [0, 0, 0, 1],
        ]
    )
    trans_inv = np.linalg.inv(trans)

    for theta in [-1 * np.pi / 6.0, 0.0, np.pi / 5.0]:  # rotation angle
        for s in [0.83, 1.3, 2.07]:  # scale
            rot = np.zeros(shape=(4, 4))
            rot[:3, :3] = geometry.rodrigues_axis_rotation(axis, theta)
            rot[3, 3] = 1.0

            scale = np.array(
                [[1 * s, 0, 0, 0], [0, 1 * s, 0, 0], [0, 0, 1 * s, 0], [0, 0, 0, 1]]
            )

            from_grid2world = trans_inv.dot(scale.dot(rot.dot(trans)))
            to_grid2world = from_grid2world.dot(scale)
            to_world2grid = np.linalg.inv(to_grid2world)

            field, assignment = vfu.create_random_displacement_3d(
                np.array(from_shape, dtype=np.int32),
                from_grid2world,
                np.array(to_shape, dtype=np.int32),
                to_grid2world,
                rng=rng,
            )
            field = np.array(field, dtype=floating)
            assignment = np.array(assignment)
            # Verify the assignments are inside the requested region
            assert_equal(0, (assignment < 0).sum())
            for i in range(3):
                assert_equal(0, (assignment[..., i] >= to_shape[i]).sum())

            # Compute the warping coordinates (see warp_2d documentation)
            Y = np.apply_along_axis(from_grid2world.dot, 0, X)[0:3, ...]
            Z = np.zeros_like(X)
            Z[0, ...] = Y[0, ...] + field[..., 0]
            Z[1, ...] = Y[1, ...] + field[..., 1]
            Z[2, ...] = Y[2, ...] + field[..., 2]
            Z[3, ...] = 1
            W = np.apply_along_axis(to_world2grid.dot, 0, Z)[0:3, ...]

            # Verify the claimed assignments are correct
            assert_array_almost_equal(W[0, ...], assignment[..., 0], 5)
            assert_array_almost_equal(W[1, ...], assignment[..., 1], 5)
            assert_array_almost_equal(W[2, ...], assignment[..., 2], 5)

    # Test exception is raised when the affine transform matrix is not valid
    valid = np.zeros((3, 4), dtype=np.float64)
    invalid = np.zeros((3, 3), dtype=np.float64)
    shape = np.array(from_shape, dtype=np.int32)
    assert_raises(
        ValueError,
        vfu.create_random_displacement_2d,
        shape,
        invalid,
        shape,
        valid,
        rng=rng,
    )
    assert_raises(
        ValueError,
        vfu.create_random_displacement_2d,
        shape,
        valid,
        shape,
        invalid,
        rng=rng,
    )


def test_harmonic_fields_2d():
    nrows = 64
    ncols = 67
    mid_row = nrows // 2
    mid_col = ncols // 2
    expected_d = np.empty(shape=(nrows, ncols, 2))
    expected_d_inv = np.empty(shape=(nrows, ncols, 2))
    for b in [0.1, 0.3, 0.7]:
        for m in [2, 4, 7]:
            for i in range(nrows):
                for j in range(ncols):
                    ii = i - mid_row
                    jj = j - mid_col
                    theta = np.arctan2(ii, jj)
                    expected_d[i, j, 0] = ii * (1.0 / (1 + b * np.cos(m * theta)) - 1.0)
                    expected_d[i, j, 1] = jj * (1.0 / (1 + b * np.cos(m * theta)) - 1.0)
                    expected_d_inv[i, j, 0] = b * np.cos(m * theta) * ii
                    expected_d_inv[i, j, 1] = b * np.cos(m * theta) * jj

            actual_d, actual_d_inv = vfu.create_harmonic_fields_2d(nrows, ncols, b, m)
            assert_array_almost_equal(expected_d, actual_d)
            assert_array_almost_equal(expected_d_inv, expected_d_inv)


def test_harmonic_fields_3d():
    nslices = 25
    nrows = 34
    ncols = 37
    mid_slice = nslices // 2
    mid_row = nrows // 2
    mid_col = ncols // 2
    expected_d = np.empty(shape=(nslices, nrows, ncols, 3))
    expected_d_inv = np.empty(shape=(nslices, nrows, ncols, 3))
    for b in [0.3, 0.7]:
        for m in [2, 5]:
            for k in range(nslices):
                for i in range(nrows):
                    for j in range(ncols):
                        kk = k - mid_slice
                        ii = i - mid_row
                        jj = j - mid_col
                        theta = np.arctan2(ii, jj)
                        expected_d[k, i, j, 0] = kk * (
                            1.0 / (1 + b * np.cos(m * theta)) - 1.0
                        )
                        expected_d[k, i, j, 1] = ii * (
                            1.0 / (1 + b * np.cos(m * theta)) - 1.0
                        )
                        expected_d[k, i, j, 2] = jj * (
                            1.0 / (1 + b * np.cos(m * theta)) - 1.0
                        )
                        expected_d_inv[k, i, j, 0] = b * np.cos(m * theta) * kk
                        expected_d_inv[k, i, j, 1] = b * np.cos(m * theta) * ii
                        expected_d_inv[k, i, j, 2] = b * np.cos(m * theta) * jj

            actual_d, actual_d_inv = vfu.create_harmonic_fields_3d(
                nslices, nrows, ncols, b, m
            )
            assert_array_almost_equal(expected_d, actual_d)
            assert_array_almost_equal(expected_d_inv, expected_d_inv)


def test_circle():
    sh = (64, 61)
    cr = sh[0] // 2
    cc = sh[1] // 2
    x_0 = np.asarray(range(sh[0]))
    x_1 = np.asarray(range(sh[1]))
    X = np.empty((2,) + sh, dtype=np.float64)
    _O = np.ones(sh)
    X[0, ...] = x_0[:, None] * _O - cr
    X[1, ...] = x_1[None, :] * _O - cc
    nrm = np.sqrt(np.sum(X**2, axis=0))
    for radius in [0, 7, 17, 32]:
        expected = nrm <= radius
        actual = vfu.create_circle(sh[0], sh[1], radius)
        assert_array_almost_equal(actual, expected)


def test_sphere():
    sh = (64, 61, 57)
    cs = sh[0] // 2
    cr = sh[1] // 2
    cc = sh[2] // 2
    x_0 = np.asarray(range(sh[0]))
    x_1 = np.asarray(range(sh[1]))
    x_2 = np.asarray(range(sh[2]))
    X = np.empty((3,) + sh, dtype=np.float64)
    _O = np.ones(sh)
    X[0, ...] = x_0[:, None, None] * _O - cs
    X[1, ...] = x_1[None, :, None] * _O - cr
    X[2, ...] = x_2[None, None, :] * _O - cc
    nrm = np.sqrt(np.sum(X**2, axis=0))
    for radius in [0, 7, 17, 32]:
        expected = nrm <= radius
        actual = vfu.create_sphere(sh[0], sh[1], sh[2], radius)
        assert_array_almost_equal(actual, expected)


def test_warping_2d():
    r"""
    Tests the cython implementation of the 2d warpings against scipy
    """
    sh = (64, 64)
    nr = sh[0]
    nc = sh[1]

    # Create an image of a circle
    radius = 24
    circle = vfu.create_circle(nr, nc, radius)
    circle = np.array(circle, dtype=floating)

    # Create a displacement field for warping
    d, dinv = vfu.create_harmonic_fields_2d(nr, nc, 0.2, 8)
    d = np.asarray(d).astype(floating)

    # Create grid coordinates
    x_0 = np.asarray(range(sh[0]))
    x_1 = np.asarray(range(sh[1]))
    X = np.empty((3,) + sh, dtype=np.float64)
    _O = np.ones(sh)
    X[0, ...] = x_0[:, None] * _O
    X[1, ...] = x_1[None, :] * _O
    X[2, ...] = 1

    # Select an arbitrary translation matrix
    t = 0.1
    trans = np.array([[1, 0, -t * nr], [0, 1, -t * nc], [0, 0, 1]])
    trans_inv = np.linalg.inv(trans)

    # Select arbitrary rotation and scaling matrices
    for theta in [-1 * np.pi / 6.0, 0.0, np.pi / 6.0]:  # rotation angle
        for s in [0.42, 1.3, 2.15]:  # scale
            ct = np.cos(theta)
            st = np.sin(theta)

            rot = np.array([[ct, -st, 0], [st, ct, 0], [0, 0, 1]])

            scale = np.array([[1 * s, 0, 0], [0, 1 * s, 0], [0, 0, 1]])

            aff = trans_inv.dot(scale.dot(rot.dot(trans)))

            # Select arbitrary (but different) grid-to-space transforms
            sampling_grid2world = scale
            field_grid2world = aff
            field_world2grid = np.linalg.inv(field_grid2world)
            image_grid2world = aff.dot(scale)
            image_world2grid = np.linalg.inv(image_grid2world)

            A = field_world2grid.dot(sampling_grid2world)
            B = image_world2grid.dot(sampling_grid2world)
            C = image_world2grid

            # Reorient the displacement field according to its grid-to-space
            # transform
            dcopy = np.copy(d)
            vfu.reorient_vector_field_2d(dcopy, field_grid2world)
            extended_dcopy = np.zeros((nr + 2, nc + 2, 2), dtype=floating)
            extended_dcopy[1 : nr + 1, 1 : nc + 1, :] = dcopy

            # Compute the warping coordinates (see warp_2d documentation)
            Y = np.apply_along_axis(A.dot, 0, X)[0:2, ...]
            Z = np.zeros_like(X)
            Z[0, ...] = map_coordinates(extended_dcopy[..., 0], Y + 1, order=1)
            Z[1, ...] = map_coordinates(extended_dcopy[..., 1], Y + 1, order=1)
            Z[2, ...] = 0
            Z = np.apply_along_axis(C.dot, 0, Z)[0:2, ...]
            T = np.apply_along_axis(B.dot, 0, X)[0:2, ...]
            W = T + Z

            # Test bilinear interpolation
            expected = map_coordinates(circle, W, order=1)
            warped = vfu.warp_2d(
                circle,
                dcopy,
                affine_idx_in=A,
                affine_idx_out=B,
                affine_disp=C,
                out_shape=np.array(sh, dtype=np.int32),
            )
            assert_array_almost_equal(warped, expected)

            # Test nearest neighbor interpolation
            expected = map_coordinates(circle, W, order=0)
            warped = vfu.warp_2d_nn(
                circle,
                dcopy,
                affine_idx_in=A,
                affine_idx_out=B,
                affine_disp=C,
                out_shape=np.array(sh, dtype=np.int32),
            )
            assert_array_almost_equal(warped, expected)

    # Test exception is raised when the affine transform matrix is not valid
    val = np.zeros((2, 3), dtype=np.float64)
    inval = np.zeros((2, 2), dtype=np.float64)
    sh = np.array(sh, dtype=np.int32)
    # Exceptions from warp_2d
    assert_raises(
        ValueError,
        vfu.warp_2d,
        circle,
        d,
        affine_idx_in=inval,
        affine_idx_out=val,
        affine_disp=val,
        out_shape=sh,
    )
    assert_raises(
        ValueError,
        vfu.warp_2d,
        circle,
        d,
        affine_idx_in=val,
        affine_idx_out=inval,
        affine_disp=val,
        out_shape=sh,
    )
    assert_raises(
        ValueError,
        vfu.warp_2d,
        circle,
        d,
        affine_idx_in=val,
        affine_idx_out=val,
        affine_disp=inval,
        out_shape=sh,
    )
    # Exceptions from warp_2d_nn
    assert_raises(
        ValueError,
        vfu.warp_2d_nn,
        circle,
        d,
        affine_idx_in=inval,
        affine_idx_out=val,
        affine_disp=val,
        out_shape=sh,
    )
    assert_raises(
        ValueError,
        vfu.warp_2d_nn,
        circle,
        d,
        affine_idx_in=val,
        affine_idx_out=inval,
        affine_disp=val,
        out_shape=sh,
    )
    assert_raises(
        ValueError,
        vfu.warp_2d_nn,
        circle,
        d,
        affine_idx_in=val,
        affine_idx_out=val,
        affine_disp=inval,
        out_shape=sh,
    )


def test_warping_3d():
    r"""
    Tests the cython implementation of the 2d warpings against scipy
    """
    sh = (64, 64, 64)
    ns = sh[0]
    nr = sh[1]
    nc = sh[2]

    # Create an image of a sphere
    radius = 24
    sphere = vfu.create_sphere(ns, nr, nc, radius)
    sphere = np.array(sphere, dtype=floating)

    # Create a displacement field for warping
    d, dinv = vfu.create_harmonic_fields_3d(ns, nr, nc, 0.2, 8)
    d = np.asarray(d).astype(floating)

    # Create grid coordinates
    x_0 = np.asarray(range(sh[0]))
    x_1 = np.asarray(range(sh[1]))
    x_2 = np.asarray(range(sh[2]))
    X = np.empty((4,) + sh, dtype=np.float64)
    _O = np.ones(sh)
    X[0, ...] = x_0[:, None, None] * _O
    X[1, ...] = x_1[None, :, None] * _O
    X[2, ...] = x_2[None, None, :] * _O
    X[3, ...] = 1

    # Select an arbitrary rotation axis
    axis = np.array([0.5, 2.0, 1.5])
    # Select an arbitrary translation matrix
    t = 0.1
    trans = np.array(
        [[1, 0, 0, -t * ns], [0, 1, 0, -t * nr], [0, 0, 1, -t * nc], [0, 0, 0, 1]]
    )
    trans_inv = np.linalg.inv(trans)

    # Select arbitrary rotation and scaling matrices
    for theta in [-1 * np.pi / 5.0, 0.0, np.pi / 5.0]:  # rotation angle
        for s in [0.45, 1.1, 2.0]:  # scale
            rot = np.zeros(shape=(4, 4))
            rot[:3, :3] = geometry.rodrigues_axis_rotation(axis, theta)
            rot[3, 3] = 1.0

            scale = np.array(
                [[1 * s, 0, 0, 0], [0, 1 * s, 0, 0], [0, 0, 1 * s, 0], [0, 0, 0, 1]]
            )

            aff = trans_inv.dot(scale.dot(rot.dot(trans)))

            # Select arbitrary (but different) grid-to-space transforms
            sampling_grid2world = scale
            field_grid2world = aff
            field_world2grid = np.linalg.inv(field_grid2world)
            image_grid2world = aff.dot(scale)
            image_world2grid = np.linalg.inv(image_grid2world)

            A = field_world2grid.dot(sampling_grid2world)
            B = image_world2grid.dot(sampling_grid2world)
            C = image_world2grid

            # Reorient the displacement field according to its grid-to-space
            # transform
            dcopy = np.copy(d)
            vfu.reorient_vector_field_3d(dcopy, field_grid2world)

            extended_dcopy = np.zeros((ns + 2, nr + 2, nc + 2, 3), dtype=floating)
            extended_dcopy[1 : ns + 1, 1 : nr + 1, 1 : nc + 1, :] = dcopy

            # Compute the warping coordinates (see warp_2d documentation)
            Y = np.apply_along_axis(A.dot, 0, X)[0:3, ...]
            Z = np.zeros_like(X)
            Z[0, ...] = map_coordinates(extended_dcopy[..., 0], Y + 1, order=1)
            Z[1, ...] = map_coordinates(extended_dcopy[..., 1], Y + 1, order=1)
            Z[2, ...] = map_coordinates(extended_dcopy[..., 2], Y + 1, order=1)
            Z[3, ...] = 0
            Z = np.apply_along_axis(C.dot, 0, Z)[0:3, ...]
            T = np.apply_along_axis(B.dot, 0, X)[0:3, ...]
            W = T + Z

            # Test bilinear interpolation
            expected = map_coordinates(sphere, W, order=1)
            warped = vfu.warp_3d(
                sphere,
                dcopy,
                affine_idx_in=A,
                affine_idx_out=B,
                affine_disp=C,
                out_shape=np.array(sh, dtype=np.int32),
            )
            assert_array_almost_equal(warped, expected, decimal=5)

            # Test nearest neighbor interpolation
            expected = map_coordinates(sphere, W, order=0)
            warped = vfu.warp_3d_nn(
                sphere,
                dcopy,
                affine_idx_in=A,
                affine_idx_out=B,
                affine_disp=C,
                out_shape=np.array(sh, dtype=np.int32),
            )
            assert_array_almost_equal(warped, expected, decimal=5)

    # Test exception is raised when the affine transform matrix is not valid
    val = np.zeros((3, 4), dtype=np.float64)
    inval = np.zeros((3, 3), dtype=np.float64)
    sh = np.array(sh, dtype=np.int32)
    # Exceptions from warp_3d
    assert_raises(
        ValueError,
        vfu.warp_3d,
        sphere,
        d,
        affine_idx_in=inval,
        affine_idx_out=val,
        affine_disp=val,
        out_shape=sh,
    )
    assert_raises(
        ValueError,
        vfu.warp_3d,
        sphere,
        d,
        affine_idx_in=val,
        affine_idx_out=inval,
        affine_disp=val,
        out_shape=sh,
    )
    assert_raises(
        ValueError,
        vfu.warp_3d,
        sphere,
        d,
        affine_idx_in=val,
        affine_idx_out=val,
        affine_disp=inval,
        out_shape=sh,
    )
    # Exceptions from warp_3d_nn
    assert_raises(
        ValueError,
        vfu.warp_3d_nn,
        sphere,
        d,
        affine_idx_in=inval,
        affine_idx_out=val,
        affine_disp=val,
        out_shape=sh,
    )
    assert_raises(
        ValueError,
        vfu.warp_3d_nn,
        sphere,
        d,
        affine_idx_in=val,
        affine_idx_out=inval,
        affine_disp=val,
        out_shape=sh,
    )
    assert_raises(
        ValueError,
        vfu.warp_3d_nn,
        sphere,
        d,
        affine_idx_in=val,
        affine_idx_out=val,
        affine_disp=inval,
        out_shape=sh,
    )


def test_affine_transforms_2d():
    r"""
    Tests 2D affine transform functions against scipy implementation
    """
    # Create a simple invertible affine transform
    d_shape = (64, 64)
    codomain_shape = (80, 80)
    nr = d_shape[0]
    nc = d_shape[1]

    # Create an image of a circle
    radius = 16
    circle = vfu.create_circle(codomain_shape[0], codomain_shape[1], radius)
    circle = np.array(circle, dtype=floating)

    # Create grid coordinates
    x_0 = np.asarray(range(d_shape[0]))
    x_1 = np.asarray(range(d_shape[1]))
    X = np.empty((3,) + d_shape, dtype=np.float64)
    _O = np.ones(d_shape)
    X[0, ...] = x_0[:, None] * _O
    X[1, ...] = x_1[None, :] * _O
    X[2, ...] = 1

    # Generate affine transforms
    t = 0.3
    trans = np.array([[1, 0, -t * nr], [0, 1, -t * nc], [0, 0, 1]])
    trans_inv = np.linalg.inv(trans)
    for theta in [-1 * np.pi / 5.0, 0.0, np.pi / 5.0]:  # rotation angle
        for s in [0.5, 1.0, 2.0]:  # scale
            ct = np.cos(theta)
            st = np.sin(theta)

            rot = np.array([[ct, -st, 0], [st, ct, 0], [0, 0, 1]])

            scale = np.array([[1 * s, 0, 0], [0, 1 * s, 0], [0, 0, 1]])

            gt_affine = trans_inv.dot(scale.dot(rot.dot(trans)))

            # Apply the affine transform to the grid coordinates
            Y = np.apply_along_axis(gt_affine.dot, 0, X)[0:2, ...]

            expected = map_coordinates(circle, Y, order=1)
            warped = vfu.transform_2d_affine(
                circle, np.array(d_shape, dtype=np.int32), affine=gt_affine
            )
            assert_array_almost_equal(warped, expected)

            # Test affine warping with nearest-neighbor interpolation
            expected = map_coordinates(circle, Y, order=0)
            warped = vfu.transform_2d_affine_nn(
                circle, np.array(d_shape, dtype=np.int32), affine=gt_affine
            )
            assert_array_almost_equal(warped, expected)

    # Test the affine = None case
    warped = vfu.transform_2d_affine(
        circle, np.array(codomain_shape, dtype=np.int32), affine=None
    )
    assert_array_equal(warped, circle)

    warped = vfu.transform_2d_affine_nn(
        circle, np.array(codomain_shape, dtype=np.int32), affine=None
    )
    assert_array_equal(warped, circle)

    # Test exception is raised when the affine transform matrix is not valid
    invalid = np.zeros((2, 2), dtype=np.float64)
    invalid_nan = np.zeros((3, 3), dtype=np.float64)
    invalid_nan[1, 1] = np.nan
    shape = np.array(codomain_shape, dtype=np.int32)
    # Exceptions from transform_2d
    assert_raises(ValueError, vfu.transform_2d_affine, circle, shape, affine=invalid)
    assert_raises(
        ValueError, vfu.transform_2d_affine, circle, shape, affine=invalid_nan
    )
    # Exceptions from transform_2d_nn
    assert_raises(ValueError, vfu.transform_2d_affine_nn, circle, shape, affine=invalid)
    assert_raises(
        ValueError, vfu.transform_2d_affine_nn, circle, shape, affine=invalid_nan
    )


def test_affine_transforms_3d():
    r"""
    Tests 3D affine transform functions against scipy implementation
    """
    # Create a simple invertible affine transform
    d_shape = (64, 64, 64)
    codomain_shape = (80, 80, 80)
    ns = d_shape[0]
    nr = d_shape[1]
    nc = d_shape[2]

    # Create an image of a sphere
    radius = 16
    sphere = vfu.create_sphere(
        codomain_shape[0], codomain_shape[1], codomain_shape[2], radius
    )
    sphere = np.array(sphere, dtype=floating)

    # Create grid coordinates
    x_0 = np.asarray(range(d_shape[0]))
    x_1 = np.asarray(range(d_shape[1]))
    x_2 = np.asarray(range(d_shape[2]))
    X = np.empty((4,) + d_shape, dtype=np.float64)
    _O = np.ones(d_shape)
    X[0, ...] = x_0[:, None, None] * _O
    X[1, ...] = x_1[None, :, None] * _O
    X[2, ...] = x_2[None, None, :] * _O
    X[3, ...] = 1

    # Generate affine transforms
    # Select an arbitrary rotation axis
    axis = np.array([0.5, 2.0, 1.5])
    t = 0.3
    trans = np.array(
        [[1, 0, 0, -t * ns], [0, 1, 0, -t * nr], [0, 0, 1, -t * nc], [0, 0, 0, 1]]
    )
    trans_inv = np.linalg.inv(trans)
    for theta in [-1 * np.pi / 5.0, 0.0, np.pi / 5.0]:  # rotation angle
        for s in [0.45, 1.1, 2.3]:  # scale
            rot = np.zeros(shape=(4, 4))
            rot[:3, :3] = geometry.rodrigues_axis_rotation(axis, theta)
            rot[3, 3] = 1.0

            scale = np.array(
                [[1 * s, 0, 0, 0], [0, 1 * s, 0, 0], [0, 0, 1 * s, 0], [0, 0, 0, 1]]
            )

            gt_affine = trans_inv.dot(scale.dot(rot.dot(trans)))

            # Apply the affine transform to the grid coordinates
            Y = np.apply_along_axis(gt_affine.dot, 0, X)[0:3, ...]

            expected = map_coordinates(sphere, Y, order=1)
            transformed = vfu.transform_3d_affine(
                sphere, np.array(d_shape, dtype=np.int32), affine=gt_affine
            )
            assert_array_almost_equal(transformed, expected)

            # Test affine transform with nearest-neighbor interpolation
            expected = map_coordinates(sphere, Y, order=0)
            transformed = vfu.transform_3d_affine_nn(
                sphere, np.array(d_shape, dtype=np.int32), affine=gt_affine
            )
            assert_array_almost_equal(transformed, expected)

    # Test the affine = None case
    transformed = vfu.transform_3d_affine(
        sphere, np.array(codomain_shape, dtype=np.int32), affine=None
    )
    assert_array_equal(transformed, sphere)

    transformed = vfu.transform_3d_affine_nn(
        sphere, np.array(codomain_shape, dtype=np.int32), affine=None
    )
    assert_array_equal(transformed, sphere)

    # Test exception is raised when the affine transform matrix is not valid
    invalid = np.zeros((3, 3), dtype=np.float64)
    invalid_nan = np.zeros((4, 4), dtype=np.float64)
    invalid_nan[1, 1] = np.nan
    shape = np.array(codomain_shape, dtype=np.int32)
    # Exceptions from transform_3d_affine
    assert_raises(ValueError, vfu.transform_3d_affine, sphere, shape, affine=invalid)
    assert_raises(
        ValueError, vfu.transform_3d_affine, sphere, shape, affine=invalid_nan
    )
    # Exceptions from transform_3d_affine_nn
    assert_raises(ValueError, vfu.transform_3d_affine_nn, sphere, shape, affine=invalid)
    assert_raises(
        ValueError, vfu.transform_3d_affine_nn, sphere, shape, affine=invalid_nan
    )


@set_random_number_generator(8315759)
def test_compose_vector_fields_2d(rng):
    r"""
    Creates two random displacement field that exactly map pixels from an input
    image to an output image. The resulting displacements and their
    composition, although operating in physical space, map the points exactly
    (up to numerical precision).
    """
    input_shape = (10, 10)
    tgt_sh = (10, 10)
    # create a simple affine transformation
    nr = input_shape[0]
    nc = input_shape[1]
    s = 1.5
    t = 2.5
    trans = np.array([[1, 0, -t * nr], [0, 1, -t * nc], [0, 0, 1]])
    trans_inv = np.linalg.inv(trans)
    scale = np.array([[1 * s, 0, 0], [0, 1 * s, 0], [0, 0, 1]])
    gt_affine = trans_inv.dot(scale.dot(trans))

    # create two random displacement fields
    input_grid2world = gt_affine
    target_grid2world = gt_affine

    disp1, assign1 = vfu.create_random_displacement_2d(
        np.array(input_shape, dtype=np.int32),
        input_grid2world,
        np.array(tgt_sh, dtype=np.int32),
        target_grid2world,
        rng=rng,
    )
    disp1 = np.array(disp1, dtype=floating)
    assign1 = np.array(assign1)

    disp2, assign2 = vfu.create_random_displacement_2d(
        np.array(input_shape, dtype=np.int32),
        input_grid2world,
        np.array(tgt_sh, dtype=np.int32),
        target_grid2world,
        rng=rng,
    )
    disp2 = np.array(disp2, dtype=floating)
    assign2 = np.array(assign2)

    # create a random image (with decimal digits) to warp
    moving_image = np.empty(tgt_sh, dtype=floating)
    moving_image[...] = rng.integers(0, 10, np.size(moving_image)).reshape(
        tuple(tgt_sh)
    )
    # set boundary values to zero so we don't test wrong interpolation due to
    # floating point precision
    moving_image[0, :] = 0
    moving_image[-1, :] = 0
    moving_image[:, 0] = 0
    moving_image[:, -1] = 0

    # evaluate the composed warping using the exact assignments
    # (first 1 then 2)
    warp1 = moving_image[(assign2[..., 0], assign2[..., 1])]
    expected = warp1[(assign1[..., 0], assign1[..., 1])]

    # compose the displacement fields
    target_world2grid = np.linalg.inv(target_grid2world)
    premult_index = target_world2grid.dot(input_grid2world)
    premult_disp = target_world2grid

    for time_scaling in [0.25, 0.5, 1.0, 2.0, 4.0]:
        composition, stats = vfu.compose_vector_fields_2d(
            disp1, disp2 / time_scaling, premult_index, premult_disp, time_scaling, None
        )
        # apply the implementation under test
        warped = np.array(
            vfu.warp_2d(
                moving_image,
                composition,
                affine_idx_in=None,
                affine_idx_out=premult_index,
                affine_disp=premult_disp,
            )
        )
        assert_array_almost_equal(warped, expected)

        # test also using nearest neighbor interpolation
        warped = np.array(
            vfu.warp_2d_nn(
                moving_image,
                composition,
                affine_idx_in=None,
                affine_idx_out=premult_index,
                affine_disp=premult_disp,
            )
        )
        assert_array_almost_equal(warped, expected)

        # test updating the displacement field instead of creating a new one
        composition = disp1.copy()
        vfu.compose_vector_fields_2d(
            composition,
            disp2 / time_scaling,
            premult_index,
            premult_disp,
            time_scaling,
            composition,
        )
        # apply the implementation under test
        warped = np.array(
            vfu.warp_2d(
                moving_image,
                composition,
                affine_idx_in=None,
                affine_idx_out=premult_index,
                affine_disp=premult_disp,
            )
        )
        assert_array_almost_equal(warped, expected)

        # test also using nearest neighbor interpolation
        warped = np.array(
            vfu.warp_2d_nn(
                moving_image,
                composition,
                affine_idx_in=None,
                affine_idx_out=premult_index,
                affine_disp=premult_disp,
            )
        )
        assert_array_almost_equal(warped, expected)

    # Test non-overlapping case
    x_0 = np.asarray(range(input_shape[0]))
    x_1 = np.asarray(range(input_shape[1]))
    X = np.empty(input_shape + (2,), dtype=np.float64)
    _O = np.ones(input_shape)
    X[..., 0] = x_0[:, None] * _O
    X[..., 1] = x_1[None, :] * _O
    random_labels = rng.integers(0, 2, input_shape[0] * input_shape[1] * 2)
    random_labels = random_labels.reshape(input_shape + (2,))
    values = np.array([-1, tgt_sh[0]])
    disp1 = (values[random_labels] - X).astype(floating)
    composition, stats = vfu.compose_vector_fields_2d(
        disp1, disp2, None, None, 1.0, None
    )
    assert_array_almost_equal(composition, np.zeros_like(composition))

    # test updating the displacement field instead of creating a new one
    composition = disp1.copy()
    vfu.compose_vector_fields_2d(composition, disp2, None, None, 1.0, composition)
    assert_array_almost_equal(composition, np.zeros_like(composition))

    # Test exception is raised when the affine transform matrix is not valid
    valid = np.zeros((2, 3), dtype=np.float64)
    invalid = np.zeros((2, 2), dtype=np.float64)
    assert_raises(
        ValueError,
        vfu.compose_vector_fields_2d,
        disp1,
        disp2,
        invalid,
        valid,
        1.0,
        None,
    )
    assert_raises(
        ValueError,
        vfu.compose_vector_fields_2d,
        disp1,
        disp2,
        valid,
        invalid,
        1.0,
        None,
    )


@set_random_number_generator(8315759)
def test_compose_vector_fields_3d(rng):
    r"""
    Creates two random displacement field that exactly map pixels from an input
    image to an output image. The resulting displacements and their
    composition, although operating in physical space, map the points exactly
    (up to numerical precision).
    """
    input_shape = (10, 10, 10)
    tgt_sh = (10, 10, 10)
    # create a simple affine transformation
    ns = input_shape[0]
    nr = input_shape[1]
    nc = input_shape[2]
    s = 1.5
    t = 2.5
    trans = np.array(
        [[1, 0, 0, -t * ns], [0, 1, 0, -t * nr], [0, 0, 1, -t * nc], [0, 0, 0, 1]]
    )
    trans_inv = np.linalg.inv(trans)
    scale = np.array(
        [[1 * s, 0, 0, 0], [0, 1 * s, 0, 0], [0, 0, 1 * s, 0], [0, 0, 0, 1]]
    )
    gt_affine = trans_inv.dot(scale.dot(trans))

    # create two random displacement fields
    input_grid2world = gt_affine
    target_grid2world = gt_affine

    disp1, assign1 = vfu.create_random_displacement_3d(
        np.array(input_shape, dtype=np.int32),
        input_grid2world,
        np.array(tgt_sh, dtype=np.int32),
        target_grid2world,
        rng=rng,
    )
    disp1 = np.array(disp1, dtype=floating)
    assign1 = np.array(assign1)

    disp2, assign2 = vfu.create_random_displacement_3d(
        np.array(input_shape, dtype=np.int32),
        input_grid2world,
        np.array(tgt_sh, dtype=np.int32),
        target_grid2world,
        rng=rng,
    )
    disp2 = np.array(disp2, dtype=floating)
    assign2 = np.array(assign2)

    # create a random image (with decimal digits) to warp
    moving_image = np.empty(tgt_sh, dtype=floating)
    moving_image[...] = rng.integers(0, 10, np.size(moving_image)).reshape(
        tuple(tgt_sh)
    )
    # set boundary values to zero so we don't test wrong interpolation due to
    # floating point precision
    moving_image[0, :, :] = 0
    moving_image[-1, :, :] = 0
    moving_image[:, 0, :] = 0
    moving_image[:, -1, :] = 0
    moving_image[:, :, 0] = 0
    moving_image[:, :, -1] = 0

    # evaluate the composed warping using the exact assignments
    # (first 1 then 2)

    warp1 = moving_image[(assign2[..., 0], assign2[..., 1], assign2[..., 2])]
    expected = warp1[(assign1[..., 0], assign1[..., 1], assign1[..., 2])]

    # compose the displacement fields
    target_world2grid = np.linalg.inv(target_grid2world)
    premult_index = target_world2grid.dot(input_grid2world)
    premult_disp = target_world2grid

    for time_scaling in [0.25, 0.5, 1.0, 2.0, 4.0]:
        composition, stats = vfu.compose_vector_fields_3d(
            disp1, disp2 / time_scaling, premult_index, premult_disp, time_scaling, None
        )
        # apply the implementation under test
        warped = np.array(
            vfu.warp_3d(
                moving_image,
                composition,
                affine_idx_in=None,
                affine_idx_out=premult_index,
                affine_disp=premult_disp,
            )
        )
        assert_array_almost_equal(warped, expected)

        # test also using nearest neighbor interpolation
        warped = np.array(
            vfu.warp_3d_nn(
                moving_image,
                composition,
                affine_idx_in=None,
                affine_idx_out=premult_index,
                affine_disp=premult_disp,
            )
        )
        assert_array_almost_equal(warped, expected)

        # test updating the displacement field instead of creating a new one
        composition = disp1.copy()
        vfu.compose_vector_fields_3d(
            composition,
            disp2 / time_scaling,
            premult_index,
            premult_disp,
            time_scaling,
            composition,
        )
        # apply the implementation under test
        warped = np.array(
            vfu.warp_3d(
                moving_image,
                composition,
                affine_idx_in=None,
                affine_idx_out=premult_index,
                affine_disp=premult_disp,
            )
        )
        assert_array_almost_equal(warped, expected)

        # test also using nearest neighbor interpolation
        warped = np.array(
            vfu.warp_3d_nn(
                moving_image,
                composition,
                affine_idx_in=None,
                affine_idx_out=premult_index,
                affine_disp=premult_disp,
            )
        )
        assert_array_almost_equal(warped, expected)

    # Test non-overlapping case
    x_0 = np.asarray(range(input_shape[0]))
    x_1 = np.asarray(range(input_shape[1]))
    x_2 = np.asarray(range(input_shape[2]))
    X = np.empty(input_shape + (3,), dtype=np.float64)
    _O = np.ones(input_shape)
    X[..., 0] = x_0[:, None, None] * _O
    X[..., 1] = x_1[None, :, None] * _O
    X[..., 2] = x_2[None, None, :] * _O
    sz = input_shape[0] * input_shape[1] * input_shape[2] * 3
    random_labels = rng.integers(0, 2, sz)
    random_labels = random_labels.reshape(input_shape + (3,))
    values = np.array([-1, tgt_sh[0]])
    disp1 = (values[random_labels] - X).astype(floating)
    composition, stats = vfu.compose_vector_fields_3d(
        disp1, disp2, None, None, 1.0, None
    )
    assert_array_almost_equal(composition, np.zeros_like(composition))

    # test updating the displacement field instead of creating a new one
    composition = disp1.copy()
    vfu.compose_vector_fields_3d(composition, disp2, None, None, 1.0, composition)
    assert_array_almost_equal(composition, np.zeros_like(composition))

    # Test exception is raised when the affine transform matrix is not valid
    valid = np.zeros((3, 4), dtype=np.float64)
    invalid = np.zeros((3, 3), dtype=np.float64)
    assert_raises(
        ValueError,
        vfu.compose_vector_fields_3d,
        disp1,
        disp2,
        invalid,
        valid,
        1.0,
        None,
    )
    assert_raises(
        ValueError,
        vfu.compose_vector_fields_3d,
        disp1,
        disp2,
        valid,
        invalid,
        1.0,
        None,
    )


def test_invert_vector_field_2d():
    r"""
    Inverts a synthetic, analytically invertible, displacement field
    """
    shape = (64, 64)
    nr = shape[0]
    nc = shape[1]
    # Create an arbitrary image-to-space transform
    t = 2.5  # translation factor

    trans = np.array([[1, 0, -t * nr], [0, 1, -t * nc], [0, 0, 1]])
    trans_inv = np.linalg.inv(trans)

    d, _ = vfu.create_harmonic_fields_2d(nr, nc, 0.2, 8)
    d = np.asarray(d).astype(floating)

    for theta in [-1 * np.pi / 5.0, 0.0, np.pi / 5.0]:  # rotation angle
        for s in [0.5, 1.0, 2.0]:  # scale
            ct = np.cos(theta)
            st = np.sin(theta)

            rot = np.array([[ct, -st, 0], [st, ct, 0], [0, 0, 1]])

            scale = np.array([[1 * s, 0, 0], [0, 1 * s, 0], [0, 0, 1]])

            gt_affine = trans_inv.dot(scale.dot(rot.dot(trans)))
            gt_affine_inv = np.linalg.inv(gt_affine)
            dcopy = np.copy(d)

            # make sure the field remains invertible after the re-mapping
            vfu.reorient_vector_field_2d(dcopy, gt_affine)

            inv_approx = vfu.invert_vector_field_fixed_point_2d(
                dcopy, gt_affine_inv, np.array([s, s]), 40, 1e-7
            )

            mapping = imwarp.DiffeomorphicMap(2, (nr, nc), disp_grid2world=gt_affine)
            mapping.forward = dcopy
            mapping.backward = inv_approx
            residual, stats = mapping.compute_inversion_error()
            assert_almost_equal(stats[1], 0, decimal=4)
            assert_almost_equal(stats[2], 0, decimal=4)

    # Test exception is raised when the affine transform matrix is not valid
    invalid = np.zeros((2, 2), dtype=np.float64)
    spacing = np.array([1.0, 1.0])
    assert_raises(
        ValueError,
        vfu.invert_vector_field_fixed_point_2d,
        d,
        invalid,
        spacing,
        40,
        1e-7,
        start=None,
    )


def test_invert_vector_field_3d():
    r"""
    Inverts a synthetic, analytically invertible, displacement field
    """
    shape = (64, 64, 64)
    ns = shape[0]
    nr = shape[1]
    nc = shape[2]

    # Create an arbitrary image-to-space transform

    # Select an arbitrary rotation axis
    axis = np.array([2.0, 0.5, 1.0])
    t = 2.5  # translation factor

    trans = np.array(
        [[1, 0, 0, -t * ns], [0, 1, 0, -t * nr], [0, 0, 1, -t * nc], [0, 0, 0, 1]]
    )
    trans_inv = np.linalg.inv(trans)

    d, _ = vfu.create_harmonic_fields_3d(ns, nr, nc, 0.2, 8)
    d = np.asarray(d).astype(floating)

    for theta in [-1 * np.pi / 5.0, 0.0, np.pi / 5.0]:  # rotation angle
        for s in [0.5, 1.0, 2.0]:  # scale
            rot = np.zeros(shape=(4, 4))
            rot[:3, :3] = geometry.rodrigues_axis_rotation(axis, theta)
            rot[3, 3] = 1.0
            scale = np.array(
                [[1 * s, 0, 0, 0], [0, 1 * s, 0, 0], [0, 0, 1 * s, 0], [0, 0, 0, 1]]
            )

            gt_affine = trans_inv.dot(scale.dot(rot.dot(trans)))
            gt_affine_inv = np.linalg.inv(gt_affine)
            dcopy = np.copy(d)

            # make sure the field remains invertible after the re-mapping
            vfu.reorient_vector_field_3d(dcopy, gt_affine)

            # Note: the spacings are used just to check convergence, so they
            # don't need to be very accurate. Here we are passing (0.5 * s) to
            # force the algorithm to make more iterations: in ANTS, there is a
            # hard-coded bound on the maximum residual, that's why we cannot
            # force more iteration by changing the parameters.
            # We will investigate this issue with more detail in the future.

            inv_approx = vfu.invert_vector_field_fixed_point_3d(
                dcopy, gt_affine_inv, np.array([s, s, s]) * 0.5, 40, 1e-7
            )

            mapping = imwarp.DiffeomorphicMap(3, (nr, nc), disp_grid2world=gt_affine)
            mapping.forward = dcopy
            mapping.backward = inv_approx
            residual, stats = mapping.compute_inversion_error()
            assert_almost_equal(stats[1], 0, decimal=3)
            assert_almost_equal(stats[2], 0, decimal=3)

    # Test exception is raised when the affine transform matrix is not valid
    invalid = np.zeros((3, 3), dtype=np.float64)
    spacing = np.array([1.0, 1.0, 1.0])
    assert_raises(
        ValueError,
        vfu.invert_vector_field_fixed_point_3d,
        d,
        invalid,
        spacing,
        40,
        1e-7,
        start=None,
    )


def test_resample_vector_field_2d():
    r"""
    Expand a vector field by 2, then subsample by 2, the resulting
    field should be the original one
    """
    domain_shape = np.array((64, 64), dtype=np.int32)
    reduced_shape = np.array((32, 32), dtype=np.int32)
    factors = np.array([0.5, 0.5])
    d, dinv = vfu.create_harmonic_fields_2d(reduced_shape[0], reduced_shape[1], 0.3, 6)
    d = np.array(d, dtype=floating)

    expanded = vfu.resample_displacement_field_2d(d, factors, domain_shape)
    subsampled = expanded[::2, ::2, :]

    assert_array_almost_equal(d, subsampled)


def test_resample_vector_field_3d():
    r"""
    Expand a vector field by 2, then subsample by 2, the resulting
    field should be the original one
    """
    domain_shape = np.array((64, 64, 64), dtype=np.int32)
    reduced_shape = np.array((32, 32, 32), dtype=np.int32)
    factors = np.array([0.5, 0.5, 0.5])
    d, dinv = vfu.create_harmonic_fields_3d(
        reduced_shape[0], reduced_shape[1], reduced_shape[2], 0.3, 6
    )
    d = np.array(d, dtype=floating)

    expanded = vfu.resample_displacement_field_3d(d, factors, domain_shape)
    subsampled = expanded[::2, ::2, ::2, :]

    assert_array_almost_equal(d, subsampled)


@set_random_number_generator(8315759)
def test_downsample_scalar_field_2d(rng):
    size = 32
    sh = (size, size)
    for reduce_r in [True, False]:
        nr = size - 1 if reduce_r else size
        for reduce_c in [True, False]:
            nc = size - 1 if reduce_c else size
            image = np.empty((size, size), dtype=floating)
            image[...] = rng.integers(0, 10, np.size(image)).reshape(sh)

            if reduce_r:
                image[-1, :] = 0
            if reduce_c:
                image[:, -1] = 0

            a = image[::2, ::2]
            b = image[1::2, ::2]
            c = image[::2, 1::2]
            d = image[1::2, 1::2]

            expected = 0.25 * (a + b + c + d)

            if reduce_r:
                expected[-1, :] *= 2
            if reduce_c:
                expected[:, -1] *= 2

            actual = np.array(vfu.downsample_scalar_field_2d(image[:nr, :nc]))
            assert_array_almost_equal(expected, actual)


@set_random_number_generator(2115556)
def test_downsample_displacement_field_2d(rng):
    size = 32
    sh = (size, size, 2)
    for reduce_r in [True, False]:
        nr = size - 1 if reduce_r else size
        for reduce_c in [True, False]:
            nc = size - 1 if reduce_c else size
            field = np.empty((size, size, 2), dtype=floating)
            field[...] = rng.integers(0, 10, np.size(field)).reshape(sh)

            if reduce_r:
                field[-1, :, :] = 0
            if reduce_c:
                field[:, -1, :] = 0

            a = field[::2, ::2, :]
            b = field[1::2, ::2, :]
            c = field[::2, 1::2, :]
            d = field[1::2, 1::2, :]

            expected = 0.25 * (a + b + c + d)

            if reduce_r:
                expected[-1, :, :] *= 2
            if reduce_c:
                expected[:, -1, :] *= 2

            actual = vfu.downsample_displacement_field_2d(field[:nr, :nc, :])
            assert_array_almost_equal(expected, actual)


@set_random_number_generator(8315759)
def test_downsample_scalar_field_3d(rng):
    size = 32
    sh = (size, size, size)
    for reduce_s in [True, False]:
        ns = size - 1 if reduce_s else size
        for reduce_r in [True, False]:
            nr = size - 1 if reduce_r else size
            for reduce_c in [True, False]:
                nc = size - 1 if reduce_c else size
                image = np.empty((size, size, size), dtype=floating)
                image[...] = rng.integers(0, 10, np.size(image)).reshape(sh)

                if reduce_s:
                    image[-1, :, :] = 0
                if reduce_r:
                    image[:, -1, :] = 0
                if reduce_c:
                    image[:, :, -1] = 0

                a = image[::2, ::2, ::2]
                b = image[1::2, ::2, ::2]
                c = image[::2, 1::2, ::2]
                d = image[1::2, 1::2, ::2]
                aa = image[::2, ::2, 1::2]
                bb = image[1::2, ::2, 1::2]
                cc = image[::2, 1::2, 1::2]
                dd = image[1::2, 1::2, 1::2]

                expected = 0.125 * (a + b + c + d + aa + bb + cc + dd)

                if reduce_s:
                    expected[-1, :, :] *= 2
                if reduce_r:
                    expected[:, -1, :] *= 2
                if reduce_c:
                    expected[:, :, -1] *= 2

                actual = vfu.downsample_scalar_field_3d(image[:ns, :nr, :nc])
                assert_array_almost_equal(expected, actual)


@set_random_number_generator(8315759)
def test_downsample_displacement_field_3d(rng):
    size = 32
    sh = (size, size, size, 3)
    for reduce_s in [True, False]:
        ns = size - 1 if reduce_s else size
        for reduce_r in [True, False]:
            nr = size - 1 if reduce_r else size
            for reduce_c in [True, False]:
                nc = size - 1 if reduce_c else size
                field = np.empty((size, size, size, 3), dtype=floating)
                field[...] = rng.integers(0, 10, np.size(field)).reshape(sh)

                if reduce_s:
                    field[-1, :, :] = 0
                if reduce_r:
                    field[:, -1, :] = 0
                if reduce_c:
                    field[:, :, -1] = 0

                a = field[::2, ::2, ::2, :]
                b = field[1::2, ::2, ::2, :]
                c = field[::2, 1::2, ::2, :]
                d = field[1::2, 1::2, ::2, :]
                aa = field[::2, ::2, 1::2, :]
                bb = field[1::2, ::2, 1::2, :]
                cc = field[::2, 1::2, 1::2, :]
                dd = field[1::2, 1::2, 1::2, :]

                expected = 0.125 * (a + b + c + d + aa + bb + cc + dd)

                if reduce_s:
                    expected[-1, :, :, :] *= 2
                if reduce_r:
                    expected[:, -1, :, :] *= 2
                if reduce_c:
                    expected[:, :, -1, :] *= 2

                actual = vfu.downsample_displacement_field_3d(field[:ns, :nr, :nc])
                assert_array_almost_equal(expected, actual)


def test_reorient_vector_field_2d():
    shape = (16, 16)
    d, dinv = vfu.create_harmonic_fields_2d(shape[0], shape[1], 0.2, 4)
    d = np.array(d, dtype=floating)

    # the vector field rotated 90 degrees
    expected = np.empty(shape=shape + (2,), dtype=floating)
    expected[..., 0] = -1 * d[..., 1]
    expected[..., 1] = d[..., 0]

    # rotate 45 degrees twice
    c = np.sqrt(0.5)
    affine = np.array([[c, -c, 0.0], [c, c, 0.0]])
    vfu.reorient_vector_field_2d(d, affine)
    vfu.reorient_vector_field_2d(d, affine)

    # verify almost equal
    assert_array_almost_equal(d, expected)

    # Test exception is raised when the affine transform matrix is not valid
    invalid = np.zeros((2, 2), dtype=np.float64)
    assert_raises(ValueError, vfu.reorient_vector_field_2d, d, invalid)


def test_reorient_vector_field_3d():
    sh = (16, 16, 16)
    d, dinv = vfu.create_harmonic_fields_3d(sh[0], sh[1], sh[2], 0.2, 4)
    d = np.array(d, dtype=floating)
    dinv = np.array(dinv, dtype=floating)

    # the vector field rotated 90 degrees around the last axis
    expected = np.empty(shape=sh + (3,), dtype=floating)
    expected[..., 0] = -1 * d[..., 1]
    expected[..., 1] = d[..., 0]
    expected[..., 2] = d[..., 2]

    # rotate 45 degrees twice around the last axis
    c = np.sqrt(0.5)
    affine = np.array([[c, -c, 0, 0], [c, c, 0, 0], [0, 0, 1, 0]])
    vfu.reorient_vector_field_3d(d, affine)
    vfu.reorient_vector_field_3d(d, affine)

    # verify almost equal
    assert_array_almost_equal(d, expected)

    # the vector field rotated 90 degrees around the first axis
    expected[..., 0] = dinv[..., 0]
    expected[..., 1] = -1 * dinv[..., 2]
    expected[..., 2] = dinv[..., 1]

    # rotate 45 degrees twice around the first axis
    affine = np.array([[1, 0, 0, 0], [0, c, -c, 0], [0, c, c, 0]])
    vfu.reorient_vector_field_3d(dinv, affine)
    vfu.reorient_vector_field_3d(dinv, affine)

    # verify almost equal
    assert_array_almost_equal(dinv, expected)

    # Test exception is raised when the affine transform matrix is not valid
    invalid = np.zeros((3, 3), dtype=np.float64)
    assert_raises(ValueError, vfu.reorient_vector_field_3d, d, invalid)


@set_random_number_generator(1134781)
def test_reorient_random_vector_fields(rng):
    # Test reorienting vector field
    for n_dims, func in (
        (2, vfu.reorient_vector_field_2d),
        (3, vfu.reorient_vector_field_3d),
    ):
        size = [20, 30, 40][:n_dims] + [n_dims]
        arr = rng.normal(size=size)
        arr_32 = arr.astype(floating)
        affine = from_matvec(rng.normal(size=(n_dims, n_dims)), np.zeros(n_dims))
        func(arr_32, affine)
        assert_almost_equal(arr_32, apply_affine(affine, arr), 6)
        # Reorient reorients without translation
        trans = np.arange(n_dims) + 2
        affine[:-1, -1] = trans
        arr_32 = arr.astype(floating)
        func(arr_32, affine)
        assert_almost_equal(arr_32, apply_affine(affine, arr) - trans, 6)

        # Test exception is raised when the affine transform is not valid
        invalid = np.eye(n_dims)
        assert_raises(ValueError, func, arr_32, invalid)


@set_random_number_generator(3921116)
def test_gradient_2d(rng):
    sh = (25, 32)
    # Create grid coordinates
    x_0 = np.asarray(range(sh[0]))
    x_1 = np.asarray(range(sh[1]))
    X = np.empty(sh + (3,), dtype=np.float64)
    _O = np.ones(sh)
    X[..., 0] = x_0[:, None] * _O
    X[..., 1] = x_1[None, :] * _O
    X[..., 2] = 1

    transform = regtransforms[("RIGID", 2)]
    theta = np.array([0.1, 5.0, 2.5])
    T = transform.param_to_matrix(theta)
    TX = X.dot(T.T)
    # Eval an arbitrary (known) function at TX
    # f(x, y) = ax^2 + bxy + cy^{2}
    # df/dx = 2ax + by
    # df/dy = 2cy + bx
    a = 2e-3
    b = 5e-3
    c = 7e-3
    img = a * TX[..., 0] ** 2 + b * TX[..., 0] * TX[..., 1] + c * TX[..., 1] ** 2
    img = img.astype(floating)
    # img is an image sampled at X with grid-to-space transform T

    # Test sparse gradient: choose some sample points (in space)
    sample = sample_domain_regular(20, np.array(sh, dtype=np.int32), T, rng=rng)
    sample = np.array(sample)
    # Compute the analytical gradient at all points
    expected = np.empty((sample.shape[0], 2), dtype=floating)
    expected[..., 0] = 2 * a * sample[:, 0] + b * sample[:, 1]
    expected[..., 1] = 2 * c * sample[:, 1] + b * sample[:, 0]
    # Get the numerical gradient with the implementation under test
    sp_to_grid = np.linalg.inv(T)
    img_spacing = np.ones(2)
    actual, inside = vfu.sparse_gradient(img, sp_to_grid, img_spacing, sample)
    diff = np.abs(expected - actual).mean(1) * inside
    # The finite differences are really not accurate, especially with float32
    assert_equal(diff.max() < 1e-3, True)
    # Verify exception is raised when passing invalid affine or spacings
    invalid_affine = np.eye(2)
    invalid_spacings = np.ones(1)
    assert_raises(
        ValueError, vfu.sparse_gradient, img, invalid_affine, img_spacing, sample
    )
    assert_raises(
        ValueError, vfu.sparse_gradient, img, sp_to_grid, invalid_spacings, sample
    )

    # Test dense gradient
    # Compute the analytical gradient at all points
    expected = np.empty(sh + (2,), dtype=floating)
    expected[..., 0] = 2 * a * TX[..., 0] + b * TX[..., 1]
    expected[..., 1] = 2 * c * TX[..., 1] + b * TX[..., 0]
    # Get the numerical gradient with the implementation under test
    sp_to_grid = np.linalg.inv(T)
    img_spacing = np.ones(2)
    actual, inside = vfu.gradient(img, sp_to_grid, img_spacing, sh, T)
    diff = np.abs(expected - actual).mean(2) * inside
    # In the dense case, we are evaluating at the exact points (sample points
    # are not slightly moved like in the sparse case) so we have more precision
    assert_equal(diff.max() < 1e-5, True)
    # Verify exception is raised when passing invalid affine or spacings
    assert_raises(ValueError, vfu.gradient, img, invalid_affine, img_spacing, sh, T)
    assert_raises(
        ValueError, vfu.gradient, img, sp_to_grid, img_spacing, sh, invalid_affine
    )
    assert_raises(ValueError, vfu.gradient, img, sp_to_grid, invalid_spacings, sh, T)


@set_random_number_generator(3921116)
def test_gradient_3d(rng):
    shape = (25, 32, 15)
    # Create grid coordinates
    x_0 = np.asarray(range(shape[0]))
    x_1 = np.asarray(range(shape[1]))
    x_2 = np.asarray(range(shape[2]))
    X = np.zeros(shape + (4,), dtype=np.float64)
    _O = np.ones(shape)
    X[..., 0] = x_0[:, None, None] * _O
    X[..., 1] = x_1[None, :, None] * _O
    X[..., 2] = x_2[None, None, :] * _O
    X[..., 3] = 1

    transform = regtransforms[("RIGID", 3)]
    theta = np.array([0.1, 0.05, 0.12, -12.0, -15.5, -7.2])
    T = transform.param_to_matrix(theta)

    TX = X.dot(T.T)
    # Eval an arbitrary (known) function at TX
    # f(x, y, z) = ax^2 + by^2 + cz^2 + dxy + exz + fyz
    # df/dx = 2ax + dy + ez
    # df/dy = 2by + dx + fz
    # df/dz = 2cz + ex + fy
    a, b, c = 2e-3, 3e-3, 1e-3
    d, e, f = 1e-3, 2e-3, 3e-3
    img = (
        a * TX[..., 0] ** 2
        + b * TX[..., 1] ** 2
        + c * TX[..., 2] ** 2
        + d * TX[..., 0] * TX[..., 1]
        + e * TX[..., 0] * TX[..., 2]
        + f * TX[..., 1] * TX[..., 2]
    )

    img = img.astype(floating)
    # Test sparse gradient: choose some sample points (in space)
    sample = sample_domain_regular(100, np.array(shape, dtype=np.int32), T, rng=rng)
    sample = np.array(sample)
    # Compute the analytical gradient at all points
    expected = np.empty((sample.shape[0], 3), dtype=floating)
    expected[..., 0] = 2 * a * sample[:, 0] + d * sample[:, 1] + e * sample[:, 2]
    expected[..., 1] = 2 * b * sample[:, 1] + d * sample[:, 0] + f * sample[:, 2]
    expected[..., 2] = 2 * c * sample[:, 2] + e * sample[:, 0] + f * sample[:, 1]
    # Get the numerical gradient with the implementation under test
    sp_to_grid = np.linalg.inv(T)
    img_spacing = np.ones(3)
    actual, inside = vfu.sparse_gradient(img, sp_to_grid, img_spacing, sample)
    # Discard points outside the image domain
    diff = np.abs(expected - actual).mean(1) * inside
    # The finite differences are really not accurate, especially with float32
    assert_equal(diff.max() < 1e-3, True)
    # Verify exception is raised when passing invalid affine or spacings
    invalid_affine = np.eye(3)
    invalid_spacings = np.ones(2)
    assert_raises(
        ValueError, vfu.sparse_gradient, img, invalid_affine, img_spacing, sample
    )
    assert_raises(
        ValueError, vfu.sparse_gradient, img, sp_to_grid, invalid_spacings, sample
    )

    # Test dense gradient
    # Compute the analytical gradient at all points
    expected = np.empty(shape + (3,), dtype=floating)
    expected[..., 0] = 2 * a * TX[..., 0] + d * TX[..., 1] + e * TX[..., 2]
    expected[..., 1] = 2 * b * TX[..., 1] + d * TX[..., 0] + f * TX[..., 2]
    expected[..., 2] = 2 * c * TX[..., 2] + e * TX[..., 0] + f * TX[..., 1]
    # Get the numerical gradient with the implementation under test
    sp_to_grid = np.linalg.inv(T)
    img_spacing = np.ones(3)
    actual, inside = vfu.gradient(img, sp_to_grid, img_spacing, shape, T)
    diff = np.abs(expected - actual).mean(3) * inside
    # In the dense case, we are evaluating at the exact points (sample points
    # are not slightly moved like in the sparse case) so we have more precision
    assert_equal(diff.max() < 1e-5, True)
    # Verify exception is raised when passing invalid affine or spacings
    assert_raises(ValueError, vfu.gradient, img, invalid_affine, img_spacing, shape, T)
    assert_raises(
        ValueError, vfu.gradient, img, sp_to_grid, img_spacing, shape, invalid_affine
    )
    assert_raises(ValueError, vfu.gradient, img, sp_to_grid, invalid_spacings, shape, T)



import os
import tempfile

import numpy as np
import pytest

from dipy.stats.sketching import count_sketch


@pytest.fixture
def setup_matrices():
    """Fixture to set up matrix A as a memmap file and return its details."""
    # Matrix A details
    matrixa_dtype = np.float64
    matrixa_shape = (100, 50)
    matrixa = np.random.rand(*matrixa_shape)

    # Create a temporary file to store matrix A
    with tempfile.NamedTemporaryFile(delete=False, suffix=".mmap") as matrixa_file:
        np.memmap(
            matrixa_file.name, dtype=matrixa_dtype, mode="w+", shape=matrixa_shape
        )[:] = matrixa
        matrixa_file_name = matrixa_file.name

    yield matrixa_file_name, matrixa_dtype, matrixa_shape

    # Cleanup: Remove the matrix file after the test
    os.unlink(matrixa_file_name)


def test_count_sketch_basic_functionality(setup_matrices):
    """Test the basic functionality of the count_sketch function."""
    matrixa_name, matrixa_dtype, matrixa_shape = setup_matrices
    sketch_rows = 10

    # Create a temporary directory for the output files
    with tempfile.TemporaryDirectory() as tmp_dir:
        matrixc_file_name, matrixc_dtype, matrixc_shape = count_sketch(
            matrixa_name, matrixa_dtype, matrixa_shape, sketch_rows, tmp_dir
        )

        # Load the resulting sketch matrix to check properties
        matrixc = np.memmap(
            matrixc_file_name, dtype=matrixc_dtype, mode="r+", shape=matrixc_shape
        )

        # Verify the output sketch matrix shape and dtype
        assert matrixc.shape == (sketch_rows, matrixa_shape[1]), "Output shape mismatch"
        assert matrixc.dtype == matrixa_dtype, "Output dtype mismatch"

        # Clean up
        del matrixc
        os.unlink(matrixc_file_name)


def test_count_sketch_randomness(setup_matrices):
    """Test if count_sketch produces different results with different random seeds."""
    matrixa_name, matrixa_dtype, matrixa_shape = setup_matrices
    sketch_rows = 10

    with tempfile.TemporaryDirectory() as tmp_dir:
        np.random.seed(42)
        matrixc_file_name_1, _, _ = count_sketch(
            matrixa_name, matrixa_dtype, matrixa_shape, sketch_rows, tmp_dir
        )
        matrixc_1 = np.memmap(matrixc_file_name_1, dtype=matrixa_dtype, mode="r+")

        np.random.seed(43)
        matrixc_file_name_2, _, _ = count_sketch(
            matrixa_name, matrixa_dtype, matrixa_shape, sketch_rows, tmp_dir
        )
        matrixc_2 = np.memmap(matrixc_file_name_2, dtype=matrixa_dtype, mode="r+")

        # Verify that the two resulting sketches are different
        assert not np.allclose(
            matrixc_1, matrixc_2
        ), "Sketches should differ with different random seeds"

        # Clean up
        del matrixc_1
        del matrixc_2
        os.unlink(matrixc_file_name_1)
        os.unlink(matrixc_file_name_2)


def test_count_sketch_preserves_shape_and_dtype(setup_matrices):
    """Test if the count_sketch preserves the shape and dtype of the input matrix A."""
    matrixa_name, matrixa_dtype, matrixa_shape = setup_matrices
    sketch_rows = 20

    with tempfile.TemporaryDirectory() as tmp_dir:
        matrixc_file_name, matrixc_dtype, matrixc_shape = count_sketch(
            matrixa_name, matrixa_dtype, matrixa_shape, sketch_rows, tmp_dir
        )

        # Check the resulting matrix properties
        assert matrixc_shape == (
            sketch_rows,
            matrixa_shape[1],
        ), "Sketch matrix shape is incorrect"
        assert matrixc_dtype == matrixa_dtype, "Sketch matrix dtype is incorrect"

        # Clean up
        os.unlink(matrixc_file_name)


def test_count_sketch_with_large_matrix():
    """Test count_sketch function with a large matrix to check performance."""
    matrixa_dtype = np.float64
    matrixa_shape = (1000, 500)
    matrixa = np.random.rand(*matrixa_shape)

    with tempfile.NamedTemporaryFile(delete=False, suffix=".mmap") as matrixa_file:
        np.memmap(
            matrixa_file.name, dtype=matrixa_dtype, mode="w+", shape=matrixa_shape
        )[:] = matrixa
        matrixa_name = matrixa_file.name

    sketch_rows = 50
    with tempfile.TemporaryDirectory() as tmp_dir:
        matrixc_file_name, matrixc_dtype, matrixc_shape = count_sketch(
            matrixa_name, matrixa_dtype, matrixa_shape, sketch_rows, tmp_dir
        )

        # Load the resulting sketch matrix
        matrixc = np.memmap(
            matrixc_file_name, dtype=matrixc_dtype, mode="r+", shape=matrixc_shape
        )

        # Verify that the sketch was successfully computed
        assert matrixc.shape == (
            sketch_rows,
            matrixa_shape[1],
        ), "Output shape mismatch for large matrix"

        # Clean up
        del matrixc
        os.unlink(matrixa_name)
        os.unlink(matrixc_file_name)
